

from .base import NbGraderPreprocessor
from .clearalwayshiddentests import ClearAlwaysHiddenTests
from .headerfooter import IncludeHeaderFooter
from .lockcells import LockCells
from .clearsolutions import ClearSolutions
from .saveautogrades import SaveAutoGrades
from .computechecksums import ComputeChecksums
from .savecells import SaveCells
from .overwritecells import OverwriteCells
from .checkcellmetadata import CheckCellMetadata
from .execute import Execute
from .getgrades import GetGrades
from .clearoutput import ClearOutput
from .limitoutput import LimitOutput
from .deduplicateids import DeduplicateIds
from .clearhiddentests import ClearHiddenTests
from .clearmarkingscheme import ClearMarkScheme
from .overwritekernelspec import OverwriteKernelspec
from .addrevert import AddRevert

__all__ = [
    "IncludeHeaderFooter",
    "LockCells",
    "ClearSolutions",
    "SaveAutoGrades",
    "ComputeChecksums",
    "SaveCells",
    "OverwriteCells",
    "CheckCellMetadata",
    "Execute",
    "GetGrades",
    "ClearOutput",
    "LimitOutput",
    "DeduplicateIds",
    "ClearHiddenTests",
    "ClearMarkScheme",
    "OverwriteKernelspec",
    "AddRevert",
    "ClearAlwaysHiddenTests"
]



import os

from grader_service.auth.auth import Authenticator
from grader_service.autograding.kube.kube_grader import KubeAutogradeExecutor
from grader_service.autograding.local_grader import LocalAutogradeExecutor
from grader_service.handlers.base_handler import BaseHandler
from grader_service.orm import User, Lecture
from grader_service.orm.base import DeleteState
from grader_service.orm.lecture import LectureState
from grader_service.orm.takepart import Scope, Role

# TODO: Only use DummyAuthenticator

# from grader_service.autograding.docker_grader import DockerAutogradeExecutor

print("### loading service config")

c.GraderService.service_host = "127.0.0.1"
# existing directory to use as the base directory for the grader service
service_dir = os.path.expanduser("~/grader_service_dir")
c.GraderService.grader_service_dir = service_dir

c.RequestHandlerConfig.autograde_executor_class = LocalAutogradeExecutor

c.CeleryApp.conf = dict(
    broker_url='amqp://localhost',
    result_backend='rpc://',
    task_serializer='json',
    result_serializer='json',
    accept_content=['json'],
    broker_connection_retry_on_startup=True,
    task_always_eager=True
)
c.CeleryApp.worker_kwargs = dict(concurrency=1, pool="prefork")


# JupyterHub client config
c.GraderService.oauth_clients = [{
    'client_id': 'my_id',
    'client_secret': 'my_secret',
    'redirect_uri': 'http://localhost:8080/hub/oauth_callback'
}]

from grader_service.auth.dummy import DummyAuthenticator

c.GraderService.authenticator_class = DummyAuthenticator
c.Authenticator.allowed_users = {'admin', 'instructor', 'student', 'tutor'}

c.Authenticator.admin_users = {'admin'}

c.GraderService.load_roles = {"lect1:instructor": ["admin", "instructor"], "lect1:student": ["student"], "lect1:tutor": ["tutor"]}



from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from grader_service.api.models.base_model import Model
from grader_service.api.models.submission import Submission
from grader_service.api.models.user import User
from grader_service.api import util

from grader_service.api.models.submission import Submission  # noqa: E501
from grader_service.api.models.user import User  # noqa: E501

class UserSubmissionsInner(Model):
    """NOTE: This class is auto generated by OpenAPI Generator (https://openapi-generator.tech).

    Do not edit the class manually.
    """

    def __init__(self, user=None, submissions=None):  # noqa: E501
        """UserSubmissionsInner - a model defined in OpenAPI

        :param user: The user of this UserSubmissionsInner.  # noqa: E501
        :type user: User
        :param submissions: The submissions of this UserSubmissionsInner.  # noqa: E501
        :type submissions: List[Submission]
        """
        self.openapi_types = {
            'user': User,
            'submissions': List[Submission]
        }

        self.attribute_map = {
            'user': 'user',
            'submissions': 'submissions'
        }

        self._user = user
        self._submissions = submissions

    @classmethod
    def from_dict(cls, dikt) -> 'UserSubmissionsInner':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The UserSubmissions_inner of this UserSubmissionsInner.  # noqa: E501
        :rtype: UserSubmissionsInner
        """
        return util.deserialize_model(dikt, cls)

    @property
    def user(self) -> User:
        """Gets the user of this UserSubmissionsInner.


        :return: The user of this UserSubmissionsInner.
        :rtype: User
        """
        return self._user

    @user.setter
    def user(self, user: User):
        """Sets the user of this UserSubmissionsInner.


        :param user: The user of this UserSubmissionsInner.
        :type user: User
        """

        self._user = user

    @property
    def submissions(self) -> List[Submission]:
        """Gets the submissions of this UserSubmissionsInner.


        :return: The submissions of this UserSubmissionsInner.
        :rtype: List[Submission]
        """
        return self._submissions

    @submissions.setter
    def submissions(self, submissions: List[Submission]):
        """Sets the submissions of this UserSubmissionsInner.


        :param submissions: The submissions of this UserSubmissionsInner.
        :type submissions: List[Submission]
        """

        self._submissions = submissions



import datetime
import json
import shutil
from http import HTTPStatus
from pathlib import Path
from typing import Union

import tornado
import os

import isodate
from grader_service.convert.gradebook.models import GradeBookModel
from grader_service.api.models.assignment import Assignment as AssignmentModel
from grader_service.api.models.assignment_settings import AssignmentSettings
from grader_service.orm.assignment import Assignment, AutoGradingBehaviour
from grader_service.orm.submission import Submission
from grader_service.orm.base import DeleteState
from grader_service.orm.takepart import Role, Scope
from grader_service.registry import VersionSpecifier, register_handler
from sqlalchemy.exc import IntegrityError
from tornado.web import HTTPError
from .handler_utils import parse_ids

from grader_service.handlers.base_handler import GraderBaseHandler, authorize


def validate_assignment_settings(settings: Union[AssignmentSettings, None]):
    if settings is None:
        return
    late_submission = settings.late_submission or []
    current_period = isodate.parse_duration("P0D")
    curr_scaling = 1.0
    for period in late_submission:
        try:
            d = isodate.parse_duration(period.period)
            if d < datetime.timedelta(0):
                raise ValueError
            if not isinstance(period.scaling, float):
                raise TypeError
            s = period.scaling
        except (isodate.isoerror.ISO8601Error, TypeError, ValueError):
            raise HTTPError(HTTPStatus.BAD_REQUEST, reason="Invalid settings!")
        if d <= current_period:
            raise HTTPError(HTTPStatus.BAD_REQUEST, reason=f"Period lengths are not increasing!)")
        if s <= 0.0 or s >= 1.0:
            raise HTTPError(HTTPStatus.BAD_REQUEST, reason=f"Score scaling has to be between 0.0 and 1.0 exclusive!")
        if s >= curr_scaling:
            raise HTTPError(HTTPStatus.BAD_REQUEST, reason=f"Score scaling is not decreasing!)")
        current_period = d
        curr_scaling = s


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/?",
    version_specifier=VersionSpecifier.ALL,
)
class AssignmentBaseHandler(GraderBaseHandler):
    """Tornado Handler for http requests

    route: /lectures/{lecture_id}/assignments."""

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int):
        """Returns all assignments of a lecture.

            :param lecture_id: id of the lecture
            :type lecture_id: int
            :raises HTTPError: throws err if lecture is deleted
            """
        lecture_id = parse_ids(lecture_id)
        self.validate_parameters("include-submissions")
        role = self.get_role(lecture_id)
        if role.lecture.deleted == DeleteState.deleted:
            raise HTTPError(HTTPStatus.NOT_FOUND, reason="Lecture not found")

        if (
                role.role == Scope.student
        ):  # students do not get assignments that are created
            assignments = (
                self.session.query(Assignment)
                .filter(
                    Assignment.lectid == role.lecture.id,
                    Assignment.deleted == DeleteState.active,
                    Assignment.status != "created",
                    Assignment.status != "pushed"
                )
                .all()
            )

        else:
            assignments = [
                a for a in role.lecture.assignments if (a.deleted
                                                        == DeleteState.active)
            ]

        # Handle the case that the user wants to include submissions
        include_submissions = self.get_argument("include-submissions", "true") == "true"
        if include_submissions:
            assignids = [a.id for a in assignments]
            username = self.user.name
            results = self.session.query(Submission).filter(
                Submission.assignid.in_(assignids),
                Submission.username == username).all()
            # Create a combined list of assignments and submissions 
            assignments = [a.serialize() for a in assignments]
            submissions = [s.serialize() for s in results]
            for a in assignments:
                a['submissions'] = [s for s in submissions if s['assignid'] == a['id']]
        assignments = sorted(assignments, key=lambda o: o["id"])
        self.write_json(assignments)

    @authorize([Scope.instructor])
    async def post(self, lecture_id: int):
        """Creates a new assignment.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        """
        lecture_id = parse_ids(lecture_id)
        self.validate_parameters()
        role = self.get_role(lecture_id)
        if role.lecture.deleted == DeleteState.deleted:
            raise HTTPError(HTTPStatus.NOT_FOUND)
        try:
            body = tornado.escape.json_decode(self.request.body)
            assignment_model = AssignmentModel.from_dict(body)
        except ValueError as e:
            # TODO Return useful error message
            raise HTTPError(HTTPStatus.BAD_REQUEST, reason=str(e))
        assignment = Assignment()

        assignment.name = assignment_model.name
        assignment_with_name = self.session.query(Assignment) \
            .filter(Assignment.name == assignment.name,
                    Assignment.deleted == DeleteState.active,
                    Assignment.lectid == lecture_id) \
            .one_or_none()

        if (assignment_with_name is not None):
            raise HTTPError(HTTPStatus.CONFLICT,
                            reason="Assignment name is already being used")
        if ((assignment_model.max_submissions is not None)
                and (assignment_model.max_submissions < 1)):
            msg = "Maximum number of submissions cannot be smaller than 1!"
            raise HTTPError(HTTPStatus.BAD_REQUEST,
                            reason=msg)
        validate_assignment_settings(assignment_model.settings)

        assignment.lectid = lecture_id
        assignment.duedate = assignment_model.due_date
        assignment.status = assignment_model.status
        assignment.type = assignment_model.type
        assignment.points = 0
        assignment.deleted = DeleteState.active
        assignment.automatic_grading = AutoGradingBehaviour.get(assignment_model.automatic_grading)
        assignment.max_submissions = assignment_model.max_submissions
        assignment.allow_files = get_allow_files(assignment_model)
        if assignment_model.settings:
            assignment.settings = json.dumps(assignment_model.settings.to_dict())

        self.session.add(assignment)
        try:
            self.session.commit()
        except IntegrityError as e:
            self.log.error(e)
            self.session.rollback()
            raise HTTPError(HTTPStatus.UNPROCESSABLE_ENTITY,
                            reason="Cannot add object to database.")
        self.set_status(HTTPStatus.CREATED)
        self.write_json(assignment)


def get_allow_files(assignment_model: AssignmentModel):
    """Extract the allow files from field in assignment table.

    Return false if field is None, else a list."""
    allow_files = False
    if assignment_model.allow_files is not None:
        allow_files = assignment_model.allow_files
    return allow_files


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/?",  # noqa E501 
    version_specifier=VersionSpecifier.ALL,
)
class AssignmentObjectHandler(GraderBaseHandler):
    """Tornado Handler for http requests

    route: /lectures/{lecture_id}/assignments/{assignment_id}."""

    @authorize([Scope.instructor])
    async def put(self, lecture_id: int, assignment_id: int):
        """Updates an assignment.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :raises HTTPError: throws err if assignment was not found
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
        self.validate_parameters()
        body = tornado.escape.json_decode(self.request.body)
        assignment_model = AssignmentModel.from_dict(body)
        assignment = self.get_assignment(lecture_id, assignment_id)
        # Validate name
        assignment_with_name = self.session.query(Assignment) \
            .filter(Assignment.name == assignment_model.name,
                    Assignment.deleted == DeleteState.active,
                    Assignment.lectid == assignment.lectid) \
            .one_or_none()

        if ((assignment_with_name is not None)
                and (assignment_with_name.id != assignment_id)):
            raise HTTPError(HTTPStatus.CONFLICT,
                            reason="Assignment name is already being used")
        if ((assignment_model.max_submissions is not None)
                and (assignment_model.max_submissions < 1)):
            raise HTTPError(HTTPStatus.BAD_REQUEST,
                            reason="Max num submissions < 1!")
        validate_assignment_settings(assignment_model.settings)

        assignment.name = assignment_model.name
        assignment.duedate = assignment_model.due_date
        assignment.status = assignment_model.status
        assignment.type = assignment_model.type
        assignment.automatic_grading = AutoGradingBehaviour[assignment_model.automatic_grading]
        assignment.max_submissions = assignment_model.max_submissions
        assignment.allow_files = get_allow_files(assignment_model)
        if assignment_model.settings:
            assignment.settings = json.dumps(assignment_model.settings.to_dict())

        if ((assignment.automatic_grading == AutoGradingBehaviour.full_auto.name)  # noqa E501
                and (assignment.properties is not None)):
            model = GradeBookModel.from_dict(json.loads(assignment.properties))
            _check_full_auto_grading(self, model)
        self.session.commit()
        self.write_json(assignment)

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int):
        """Returns a specific assignment of a lecture.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :raises HTTPError: throws err if assignment was not found
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
        assignment = self.get_assignment(lecture_id=lecture_id, assignment_id=assignment_id)
        self.write_json(assignment)

    @authorize([Scope.instructor])
    async def delete(self, lecture_id: int, assignment_id: int):
        """Soft-Deletes a specific assignment.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :raises HTTPError: throws err if assignment was not found or deleted
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
        self.validate_parameters()
        assignment = self.get_assignment(lecture_id, assignment_id)

        if assignment.status in ["released", "complete"]:
            msg = "Cannot delete released or completed assignments!"
            raise HTTPError(HTTPStatus.CONFLICT,
                            reason=msg)

        previously_deleted = (
            self.session.query(Assignment)
            .filter(
                Assignment.lectid == lecture_id,
                Assignment.name == assignment.name,
                Assignment.deleted == DeleteState.deleted,
            )
            .one_or_none()
        )
        if previously_deleted is not None:
            self.session.delete(previously_deleted)
            self.session.commit()

        assignment.deleted = DeleteState.deleted
        self.session.commit()


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/reset\/?",  # noqa E501
    version_specifier=VersionSpecifier.ALL,
)
class AssignmentResetHandler(GraderBaseHandler):
    """Tornado Handler for http requests

    route: /lectures/{lecture_id}/assignments/{assignment_id}/reset."""

    @authorize([Scope.instructor, Scope.tutor, Scope.student])
    async def get(self, lecture_id: int, assignment_id: int):
        self.validate_parameters()
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
        assignment = self.get_assignment(lecture_id, assignment_id)

        grader_service_dir = Path(self.application.grader_service_dir)
        git_path_base = grader_service_dir.joinpath(
            "tmp",
            assignment.lecture.code,
            assignment.name,
            self.user.name)
        self.log.info(git_path_base)
        # Deleting dir
        if os.path.exists(git_path_base):
            shutil.rmtree(git_path_base)

        self.log.info(f"DIR {git_path_base}")
        os.makedirs(git_path_base, exist_ok=True)
        git_path_release = os.path.join(git_path_base, "release")
        git_path_user = os.path.join(git_path_base, self.user.name)
        self.log.info(f"GIT BASE {git_path_base}")
        self.log.info(f"GIT RELEASE {git_path_release}")
        self.log.info(f"GIT USER {git_path_user}")

        repo_path_release = self.construct_git_dir('release',
                                                   assignment.lecture,
                                                   assignment)
        repo_path_user = self.construct_git_dir(assignment.type,
                                                assignment.lecture,
                                                assignment)

        self.duplicate_release_repo(repo_path_release=repo_path_release,
                                    repo_path_user=repo_path_user,
                                    assignment=assignment,
                                    message="Reset Assignment")

        self.write_json(assignment)


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/properties\/?",  # noqa E501
    version_specifier=VersionSpecifier.ALL,
)
class AssignmentPropertiesHandler(GraderBaseHandler):
    """Tornado Handler for http requests

    route: /lectures/{lecture_id}/assignments/{assignment_id}/properties."""

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int):
        """
        Returns the properties of a specific assignment.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :raises HTTPError: if assignment or their properties not found
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)

        self.validate_parameters()
        assignment = self.get_assignment(lecture_id, assignment_id)
        if assignment.properties is not None:
            if self.get_role(lecture_id).role == Scope.student:
                model = GradeBookModel.from_dict(
                    json.loads(assignment.properties))
                for notebook in model.notebooks.values():
                    notebook.source_cells_dict = {}
                self.write(json.dumps(model.to_dict()))
            else:
                self.write(assignment.properties)
        else:
            raise HTTPError(HTTPStatus.NOT_FOUND,
                            reason="Assignment not found")

    @authorize([Scope.tutor, Scope.instructor])
    async def put(self, lecture_id: int, assignment_id: int):
        """
        Updates the properties of a specific assignment.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :raises HTTPError: throws err if the assignment was not found
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
        self.validate_parameters()
        assignment = self.get_assignment(lecture_id, assignment_id)
        properties_string: str = self.request.body.decode("utf-8")

        model = GradeBookModel.from_dict(json.loads(properties_string))
        # Check if assignment contains no cells that
        # need manual grading if assignment is fully auto graded
        if assignment.automatic_grading == AutoGradingBehaviour.full_auto:
            _check_full_auto_grading(self, model)

        assignment.properties = properties_string
        assignment.points = model.max_score
        self.session.commit()


def _check_full_auto_grading(self: GraderBaseHandler, model):
    """Check if the assignment notebook contains manually graded cells

    If cells are found, throw error.
    :param self: handler class
    :param model: the notebook which is being tested
    :return: void"""
    for nb in model.notebooks.values():
        if len(nb.task_cells_dict) > 0:
            msg = "Fully autograded notebook cannot contain task cells!"
            raise HTTPError(HTTPStatus.CONFLICT,
                            reason=msg)
        grades = set(nb.grade_cells_dict.keys())
        solutions = set(nb.solution_cells_dict.keys())
        if len(grades & solutions) > 0:
            msg = "Fully autograded notebook cannot \
                    contain manually graded cells!"
            raise HTTPError(HTTPStatus.CONFLICT,
                            reason=msg)





from .base import BaseConverter, GraderConvertException
from .generate_assignment import GenerateAssignment
from .autograde import Autograde
from .generate_feedback import GenerateFeedback
from .generate_solution import GenerateSolution

__all__ = [
    "BaseConverter",
    "GenerateAssignment",
    "Autograde",
    "GenerateFeedback",
    "GenerateSolution"
]



import json
from base64 import encodebytes, decodebytes

from sqlalchemy import TypeDecorator
from sqlalchemy.types import Text
from tornado.log import app_log


class JSONDict(TypeDecorator):
    """Represents an immutable structure as a json-encoded string.

    Usage::

        JSONDict(255)

    """

    impl = Text

    def _json_default(self, obj):
        """encode non-jsonable objects as JSON

        Currently only bytes are supported

        """
        if not isinstance(obj, bytes):
            app_log.warning(
                "Non-jsonable data in user_options: %r; will persist None.", type(obj)
            )
            return None

        return {"__jupyterhub_bytes__": True, "data": encodebytes(obj).decode('ascii')}

    def _object_hook(self, dct):
        """decode non-json objects packed by _json_default"""
        if dct.get("__jupyterhub_bytes__", False):
            return decodebytes(dct['data'].encode('ascii'))
        return dct

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value, default=self._json_default)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value, object_hook=self._object_hook)
        return value


class JSONList(JSONDict):
    """Represents an immutable structure as a json-encoded string (to be used for list type columns).

    Usage::

        JSONList(JSONDict)

    """

    def process_bind_param(self, value, dialect):
        if isinstance(value, list) and value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is None:
            return []
        else:
            value = json.loads(value)
        return value




from textwrap import dedent

from grader_service.convert.converters.autograde import AutogradeApp
from grader_service.convert.converters.baseapp import (
    ConverterApp,
    base_converter_aliases,
    base_converter_flags,
)
from grader_service.convert.converters.generate_assignment import GenerateAssignmentApp
from grader_service.convert.converters.generate_feedback import GenerateFeedbackApp
from grader_service.convert.converters.generate_solution import GenerateSolutionApp
from grader_service.convert.converters.updateapp import UpdateApp
from grader_service.convert.converters.validateapp import ValidateApp

aliases = {}
aliases.update(base_converter_aliases)
aliases.update({})

flags = {}
flags.update(base_converter_flags)
flags.update({})


class GraderConverter(ConverterApp):

    name = u"grader-converter"
    description = u"Convert notebooks to different formats"
    version = ConverterApp.__version__

    aliases = aliases
    flags = flags

    subcommands = dict(
        generate_assignment=(
            GenerateAssignmentApp,
            dedent(
                """
                Create the student version of an assignment. Intended for use by
                instructors only.
                """
            ).strip(),
        ),
        generate_solution=(
            GenerateSolutionApp,
            dedent(
                """
                Generates the solution for the given assignment.
                """
            ).strip(),
        ),
        autograde=(
            AutogradeApp,
            dedent(
                """
                Autograde submitted assignments. Intended for use by instructors
                only.
                """
            ).strip(),
        ),
        generate_feedback=(
            GenerateFeedbackApp,
            dedent(
                """
                Generate feedback (after autograding and manual grading).
                Intended for use by instructors only.
                """
            ).strip(),
        ),
        validate=(
            ValidateApp,
            dedent(
                """
                Validate a notebook. Intended for use by instructors and students.
                """
            ).strip(),
        ),
        update=(
            UpdateApp,
            dedent(
                """
                Update nbgrader cell metadata to the most recent version.
                """
            ).strip(),
        ),
    )


def main():
    GraderConverter.launch_instance()


if __name__ == "__main__":
    main()




from sys import path
from textwrap import dedent
from typing import Any

from grader_service.convert.converters.base import BaseConverter
from grader_service.convert.converters.baseapp import ConverterApp
from grader_service.convert.preprocessors import (
    AddRevert,
    CheckCellMetadata,
    ClearHiddenTests,
    ClearMarkScheme,
    ClearOutput,
    ClearSolutions,
    ComputeChecksums,
    IncludeHeaderFooter,
    LockCells,
    SaveCells,
    ClearAlwaysHiddenTests
)
from traitlets import Bool, List, default
from traitlets.config.loader import Config

class GenerateAssignment(BaseConverter):
    create_assignment = Bool(
        True,
        help=dedent(
            """
            Whether to create the assignment at runtime if it does not
            already exist.
            """
        ),
    ).tag(config=True)

    @default("permissions")
    def _permissions_default(self) -> int:
        return 664

    preprocessors = List(
        [
            IncludeHeaderFooter,
            LockCells,
            ClearSolutions,
            ClearOutput,
            CheckCellMetadata,
            ComputeChecksums,
            SaveCells,
            ClearHiddenTests,
            ClearAlwaysHiddenTests,
            ClearMarkScheme,
            ComputeChecksums,
            AddRevert,
            CheckCellMetadata,
        ]
    ).tag(config=True)

    # NB: ClearHiddenTests must come after ComputeChecksums and SaveCells.
    # ComputerChecksums must come again after ClearHiddenTests.

    def _load_config(self, cfg: Config, **kwargs: Any) -> None:
        super(GenerateAssignment, self)._load_config(cfg, **kwargs)

    def __init__(
            self, input_dir: str, output_dir: str, file_pattern: str, **kwargs: Any
    ) -> None:
        super(GenerateAssignment, self).__init__(
            input_dir, output_dir, file_pattern, **kwargs
        )
        self.force = True  # always overwrite generated assignments

    def start(self) -> None:
        super(GenerateAssignment, self).start()


class GenerateAssignmentApp(ConverterApp):
    version = ConverterApp.__version__

    def start(self):
        GenerateAssignment(
            input_dir=self.input_directory,
            output_dir=self.output_directory,
            file_pattern=self.file_pattern,
            copy_files=self.copy_files,
            config=self.config
        ).start()



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import enum
from typing import List, Tuple

from tornado.web import RequestHandler


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args,
                                                                 **kwargs)
        return cls._instances[cls]


class HandlerPathRegistry(object, metaclass=Singleton):
    registry = {}

    @staticmethod
    def handler_list(
            base_url: str = "/"
    ) -> List[Tuple[str, RequestHandler]]:
        return list(
            zip(
                [
                    base_url.rstrip('/').replace("/", "\\/") + path
                    for path in HandlerPathRegistry.registry.values()
                ],
                HandlerPathRegistry.registry.keys(),
            )
        )

    @staticmethod
    def has_path(cls) -> bool:
        return cls in HandlerPathRegistry.registry

    @staticmethod
    def get_path(cls):
        return HandlerPathRegistry.registry[cls]

    @staticmethod
    def add(cls, path: str):
        # check if class inherits from tornado RequestHandler
        if RequestHandler not in cls.__mro__:
            err = "Incorrect base class. "
            err += "Class must be extended from tornado 'RequestHandler' "
            err += "to be registered."
            raise ValueError(err)
        HandlerPathRegistry.registry[cls] = path


class VersionSpecifier(enum.Enum):
    ALL = "all"
    NONE = "none"
    V1 = "1"


def register_handler(
        path: str, version_specifier: VersionSpecifier = VersionSpecifier.NONE
):
    # TODO
    # add optional /services/grader prefix regex
    # any endpoint can also be accessed through /services/grader/<endpoint> 
    # this enables the service to hide behind a jupyterhub as a managed service
    
     
    # add version specifier if set
    if version_specifier == VersionSpecifier.ALL:
        # only supports single digit versions
        excludes = ["all", "none"]
        values = [v.value for v in VersionSpecifier if v.value not in excludes]
        regex_versions = "".join(values)
        v = r"(?:\/v[{}])?".format(regex_versions)
    elif ((version_specifier == VersionSpecifier.NONE)
            or (version_specifier is None)):
        v = ""
    else:
        v = rf"\/v{version_specifier.value}"
    path = v + path

    def _register_class(cls):
        HandlerPathRegistry().add(cls, path)
        return cls

    return _register_class



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

"""initial

Revision ID: a1791b1371ed
Revises: 
Create Date: 2021-05-05 11:42:24.126371

"""
from datetime import datetime

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql.schema import UniqueConstraint

# revision identifiers, used by Alembic.
revision = 'a1791b1371ed'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('lecture',
                    sa.Column('id', sa.Integer(),
                              autoincrement=True, nullable=False),

                    sa.Column('name', sa.String(length=255), nullable=True),

                    sa.Column('code', sa.String(length=255),
                              nullable=True, unique=True),

                    sa.Column('state', sa.Enum('active', 'complete',
                                               name='lecture_state'),
                              nullable=False),

                    sa.Column('deleted', sa.Enum('active', 'deleted',
                                                 name='deleted'),
                              server_default='active',
                              nullable=False),

                    sa.Column('created_at', sa.DateTime(),
                              default=datetime.utcnow, nullable=False),

                    sa.Column('updated_at', sa.DateTime(),
                              default=datetime.utcnow,
                              onupdate=datetime.utcnow,
                              nullable=False),

                    sa.PrimaryKeyConstraint('id')
                    )

    op.create_table('user',
                    sa.Column('name', sa.String(length=255), nullable=False),
                    sa.PrimaryKeyConstraint('name')
                    )

    op.create_table('assignment',
                    sa.Column('id', sa.Integer(),
                              autoincrement=True, nullable=False),

                    sa.Column('name', sa.String(length=255), nullable=False),

                    sa.Column('type', sa.Enum("user", "group",
                                              name='assignment_type'),
                              nullable=False,
                              server_default="user"),

                    sa.Column('lectid', sa.Integer(), nullable=True),

                    sa.Column('duedate', sa.DateTime(), nullable=True),

                    sa.Column('automatic_grading',
                              sa.Enum('unassisted', 'auto',
                                      'full_auto', name='automatic_grading'),
                              server_default='unassisted', nullable=False),

                    sa.Column('points', sa.DECIMAL(10, 3), nullable=False),

                    sa.Column('status',
                              sa.Enum('created', 'pushed', 'released',
                                      'fetching', 'fetched', 'complete',
                                      name='assignment_status'),
                              nullable=True),

                    sa.Column('deleted',
                              sa.Enum('active', 'deleted', name='deleted'),
                              server_default='active',
                              nullable=False),

                    sa.Column('max_submissions', sa.Integer(),
                              nullable=True, default=None, unique=False),

                    sa.Column('allow_files', sa.Boolean(),
                              nullable=False, default=False),
                    sa.Column('properties', sa.Text(),
                              nullable=True, unique=False),

                    sa.Column('created_at', sa.DateTime(),
                              default=datetime.utcnow, nullable=False),

                    sa.Column('updated_at', sa.DateTime(),
                              default=datetime.utcnow,
                              onupdate=datetime.utcnow,
                              nullable=False),

                    sa.ForeignKeyConstraint(['lectid'],
                                            ['lecture.id'], ),
                    sa.PrimaryKeyConstraint('id'),
                    UniqueConstraint('name', 'lectid',
                                     'deleted', name='u_name_in_lect')
                    )

    op.create_table('takepart',
                    sa.Column('username', sa.String(length=255),
                              nullable=False),
                    sa.Column('lectid', sa.Integer(), nullable=False),
                    sa.Column('role', sa.String(length=255), nullable=False),
                    sa.ForeignKeyConstraint(['lectid'], ['lecture.id'], ),
                    sa.ForeignKeyConstraint(['username'], ['user.name'], ),
                    sa.PrimaryKeyConstraint('username', 'lectid')
                    )

    op.create_table('group',
                    sa.Column('id', sa.Integer(), autoincrement=True),
                    sa.Column('name', sa.String(length=255), nullable=False),
                    sa.Column('lectid', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(['lectid'], ['lecture.id'], ),
                    sa.PrimaryKeyConstraint('id')
                    )
    op.create_table('partof',
                    sa.Column('username', sa.String(length=255), nullable=False),
                    sa.Column('group_id', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(['group_id'], ['group.id'], ),
                    sa.ForeignKeyConstraint(['username'], ['user.name'], ),
                    sa.PrimaryKeyConstraint('username', 'group_id')
                    )

    op.create_table('submission',
                    sa.Column('id', sa.Integer(), autoincrement=True,
                              nullable=False),

                    sa.Column('date', sa.DateTime(), nullable=True),

                    sa.Column('auto_status',
                              sa.Enum('pending', 'not_graded',
                                      'automatically_graded',
                                      'grading_failed', name='auto_status'),
                              nullable=False),

                    sa.Column('manual_status',
                              sa.Enum('not_graded', 'manually_graded',
                                      'being_edited', name='manual_status'),
                              nullable=False),

                    sa.Column('edited', sa.Boolean(), default=False),

                    sa.Column('score', sa.DECIMAL(10, 3), nullable=True),

                    sa.Column('assignid', sa.Integer(), nullable=True),

                    sa.Column('username', sa.String(length=255),
                              nullable=True),

                    sa.Column('commit_hash', sa.String(length=40),
                              nullable=False),

                    sa.Column('feedback_available', sa.Boolean(),
                              nullable=False, server_default="false"),

                    sa.Column('updated_at', sa.DateTime(),
                              default=datetime.utcnow,
                              onupdate=datetime.utcnow,
                              nullable=False),

                    sa.ForeignKeyConstraint(['assignid'], ['assignment.id'], ),
                    sa.ForeignKeyConstraint(['username'], ['user.name'], ),
                    sa.PrimaryKeyConstraint('id')
                    )

    op.create_table('submission_logs',
                    sa.Column('sub_id', sa.Integer(), nullable=False),

                    sa.Column('logs', sa.Text(), nullable=True),

                    sa.ForeignKeyConstraint(['sub_id'], ['submission.id'], ),
                    sa.PrimaryKeyConstraint('sub_id')
                    )

    op.create_table('submission_properties',
                    sa.Column('sub_id', sa.Integer(), nullable=False),

                    sa.Column('properties', sa.Text(),
                              nullable=True, unique=False),

                    sa.ForeignKeyConstraint(['sub_id'], ['submission.id'], ),
                    sa.PrimaryKeyConstraint('sub_id')
                    )

    # ### end Alembic commands ###


def downgrade():
    # This does not drop enum types which are required when using postgresql
    op.drop_table('submission_logs')
    op.drop_table('submission_properties')
    op.drop_table('submission')
    op.drop_table('takepart')
    op.drop_table('assignment')
    op.drop_table('partof')
    op.drop_table('group')
    op.drop_table('lecture')
    op.drop_table('user')
    # ### end Alembic commands ###



from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from grader_service.api.models.base_model import Model
from grader_service.api import util


class RemoteFileStatus(Model):
    """NOTE: This class is auto generated by OpenAPI Generator (https://openapi-generator.tech).

    Do not edit the class manually.
    """

    def __init__(self, status=None):  # noqa: E501
        """RemoteFileStatus - a model defined in OpenAPI

        :param status: The status of this RemoteFileStatus.  # noqa: E501
        :type status: str
        """
        self.openapi_types = {
            'status': str
        }

        self.attribute_map = {
            'status': 'status'
        }

        self._status = status

    @classmethod
    def from_dict(cls, dikt) -> 'RemoteFileStatus':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The RemoteFileStatus of this RemoteFileStatus.  # noqa: E501
        :rtype: RemoteFileStatus
        """
        return util.deserialize_model(dikt, cls)

    @property
    def status(self) -> str:
        """Gets the status of this RemoteFileStatus.


        :return: The status of this RemoteFileStatus.
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status: str):
        """Sets the status of this RemoteFileStatus.


        :param status: The status of this RemoteFileStatus.
        :type status: str
        """
        allowed_values = ["UP_TO_DATE", "DIVERGENT", "PULL_NEEDED", "PUSH_NEEDED", "NO_REMOTE_REPO"]  # noqa: E501
        if status not in allowed_values:
            raise ValueError(
                "Invalid value for `status` ({0}), must be one of {1}"
                .format(status, allowed_values)
            )

        self._status = status



"""add_submission_softdelete_column

Revision ID: a0718dae969d
Revises: 9cfeb0faa0c0
Create Date: 2024-07-06 12:12:59.091247

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'a0718dae969d'
down_revision = 'ba71c755c153'
branch_labels = None
depends_on = None


def upgrade():
   op.add_column('submission', 
                  sa.Column('deleted', 
                            sa.Enum('active', 'deleted', name='deleted'), 
                            server_default='active', 
                            nullable=False))
    
   op.execute("UPDATE submission SET deleted = 'active'")


def downgrade():
    op.drop_column('submission', 'deleted')



from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from grader_service.api.models.base_model import Model
from grader_service.api.models.submission_period import SubmissionPeriod
from grader_service.api import util

from grader_service.api.models.submission_period import SubmissionPeriod  # noqa: E501

class AssignmentSettings(Model):
    """NOTE: This class is auto generated by OpenAPI Generator (https://openapi-generator.tech).

    Do not edit the class manually.
    """

    def __init__(self, late_submission=None):  # noqa: E501
        """AssignmentSettings - a model defined in OpenAPI

        :param late_submission: The late_submission of this AssignmentSettings.  # noqa: E501
        :type late_submission: List[SubmissionPeriod]
        """
        self.openapi_types = {
            'late_submission': List[SubmissionPeriod]
        }

        self.attribute_map = {
            'late_submission': 'late_submission'
        }

        self._late_submission = late_submission

    @classmethod
    def from_dict(cls, dikt) -> 'AssignmentSettings':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The AssignmentSettings of this AssignmentSettings.  # noqa: E501
        :rtype: AssignmentSettings
        """
        return util.deserialize_model(dikt, cls)

    @property
    def late_submission(self) -> List[SubmissionPeriod]:
        """Gets the late_submission of this AssignmentSettings.


        :return: The late_submission of this AssignmentSettings.
        :rtype: List[SubmissionPeriod]
        """
        return self._late_submission

    @late_submission.setter
    def late_submission(self, late_submission: List[SubmissionPeriod]):
        """Sets the late_submission of this AssignmentSettings.


        :param late_submission: The late_submission of this AssignmentSettings.
        :type late_submission: List[SubmissionPeriod]
        """

        self._late_submission = late_submission



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from sqlalchemy import Column, ForeignKey, Integer, String, Table
from sqlalchemy.orm import relationship

from grader_service.orm.base import Base

group_assignment_table = Table('partof', Base.metadata,
                               Column('username', ForeignKey('user.name'),
                                      primary_key=True),
                               Column('group_name', ForeignKey('group.name'),
                                      primary_key=True)
                               )


class Group(Base):
    __tablename__ = "group"
    name = Column(String(255))
    lectid = Column(Integer, ForeignKey("lecture.id"), primary_key=True)

    lecture = relationship("Lecture", back_populates="groups")
    users = relationship("User", secondary=group_assignment_table,
                         back_populates="groups")





from textwrap import dedent
from typing import Tuple

from nbconvert.exporters.exporter import ResourcesDict
from nbformat.notebooknode import NotebookNode
from traitlets import Bool, Unicode

from grader_service.convert import utils
from grader_service.convert.preprocessors.base import NbGraderPreprocessor


class ClearAlwaysHiddenTests(NbGraderPreprocessor):

    begin_util_delimeter = Unicode(
        "BEGIN ALWAYS HIDDEN TESTS",
        help="The delimiter marking the beginning of hidden util code",
    ).tag(config=True)

    end_util_delimeter = Unicode(
        "END ALWAYS HIDDEN TESTS", help="The delimiter marking the end of hidden tests cases"
    ).tag(config=True)

    enforce_metadata = Bool(
        True,
        help=dedent(
            """
            Whether or not to complain if cells containing hidden util regions
            are not marked as grade cells.
            """
        ),
    ).tag(config=True)

    def _remove_hidden_util_region(self, cell: NotebookNode) -> bool:
        """Find a region in the cell that is delimeted by
        `self.begin_util_delimeter` and `self.end_util_delimeter` (e.g.  ###
        BEGIN HIDDEN UTILS and ### END HIDDEN UTILS). Remove that region
        depending on the cell type.

        This modifies the cell in place, and then returns True if a
        hidden test region was removed, and False otherwise.
        """
        # pull out the cell input/source
        lines = cell.source.split("\n")

        new_lines = []
        in_util = False
        removed_util = False

        for line in lines:
            # begin the test area
            if self.begin_util_delimeter in line:

                # check to make sure this isn't a nested BEGIN ALWAYS HIDDEN UTILS
                # region
                if in_util:
                    raise RuntimeError(
                        "Encountered nested begin always hidden statements"
                    )
                in_util = True
                removed_util = True

            # end the utils area
            elif self.end_util_delimeter in line:
                in_util = False

            # add lines as long as it's not in the hidden util region
            elif not in_util:
                new_lines.append(line)

        # we finished going through all the lines, but didn't find a
        # matching END HIDDEN UTILS statment
        if in_util:
            raise RuntimeError("No end always hidden utils statement found")

        # replace the cell source
        cell.source = "\n".join(new_lines)

        return removed_util

    def preprocess(
        self, nb: NotebookNode, resources: ResourcesDict
    ) -> Tuple[NotebookNode, ResourcesDict]:
        nb, resources = super(ClearAlwaysHiddenTests, self).preprocess(nb, resources)
        if "celltoolbar" in nb.metadata:
            del nb.metadata["celltoolbar"]
        return nb, resources

    def preprocess_cell(
        self, cell: NotebookNode, resources: ResourcesDict, cell_index: int
    ) -> Tuple[NotebookNode, ResourcesDict]:
        # remove hidden test regions
        removed_util = self._remove_hidden_util_region(cell)

        # determine whether the cell is a grade cell
        is_grade = utils.is_grade(cell)

        # check that it is marked as a grade cell if we remove a test
        # region -- if it's not, then this is a problem, because the cell needs
        # to be given an id
        if not is_grade and removed_util:
            if self.enforce_metadata:
                raise RuntimeError(
                    "Hidden util region detected in a non-grade cell; "
                    "please make sure all solution regions are within "
                    "'Autograded test cells' cells."
                )

        return cell, resources



"""Added oauth provider functionality

Revision ID: ba71c755c153
Revises: 9cfeb0faa0c0
Create Date: 2023-08-18 10:30:24.465356

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy import Column, Unicode, Integer, DateTime, Text, \
    PrimaryKeyConstraint, ForeignKeyConstraint

from grader_service.utils import new_token

# revision identifiers, used by Alembic.
revision = 'ba71c755c153'
down_revision = '9cfeb0faa0c0'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table('api_token',
                    Column('username', Unicode(255)),
                    Column('id', Integer, nullable=True),
                    Column('hashed', Unicode(255)),
                    Column('prefix', Unicode(16)),
                    Column('client_id', Unicode(255)),
                    Column('session_id', Unicode(255)),
                    Column('created', DateTime),
                    Column('expires_at', DateTime),
                    Column('last_activity', DateTime),
                    Column('note', Unicode(1023)),
                    Column('scopes', Text),
                    PrimaryKeyConstraint('id'),
                    ForeignKeyConstraint(['username'], ['user.name'])
                    )

    op.create_table('oauth_client',
                    Column('id', Integer, nullable=True),
                    Column('identifier', Unicode(255), unique=True),
                    Column('description', Unicode(1023)),
                    Column('secret', Unicode(255)),
                    Column('redirect_uri', Unicode(1023)),
                    Column('allowed_scopes', Text),
                    PrimaryKeyConstraint('id')
                    )

    op.create_table('oauth_code',
                    Column('id', Integer, nullable=True),
                    Column('client_id', Unicode(255)),
                    Column('code', Unicode(36)),
                    Column('expires_at', Integer),
                    Column('redirect_uri', Unicode(1023)),
                    Column('session_id', Unicode(255)),
                    Column('username', Unicode(255)),
                    Column('scopes', Text),
                    PrimaryKeyConstraint("id"),
                    ForeignKeyConstraint(['client_id'],
                                         ['oauth_client.identifier']),
                    ForeignKeyConstraint(['username'], ['user.name'])
                    )

    op.add_column('user', sa.Column('encrypted_auth_state', sa.types.LargeBinary, nullable=True))
    op.add_column('user', sa.Column('cookie_id', Unicode(255), nullable=True))

    connection = op.get_bind()
    result = connection.execute(sa.text('select * from "user"'))
    for row in result:
        connection.execute(sa.text('UPDATE "user" SET cookie_id = :uuid WHERE name = :name'), uuid=new_token(), name=row["name"])

    if connection.dialect.name != "sqlite":
        op.alter_column('user', 'cookie_id', nullable=False)
        op.create_unique_constraint('uq_user_cookie', 'user', ['cookie_id'])
    else:
        with op.batch_alter_table('user') as batch_op:
            batch_op.alter_column('cookie_id', nullable=False)
            batch_op.create_unique_constraint('uq_user_cookie', ['cookie_id'])


def downgrade():
    connection = op.get_bind()
    op.drop_table('api_token')
    op.drop_table('oauth_code')
    op.drop_table('oauth_client')
    op.drop_column('user', 'encrypted_auth_state')
    if connection.dialect.name != "sqlite":
        op.drop_constraint('uq_user_cookie', 'user')
    else:
        with op.batch_alter_table('user') as batch_op:
            batch_op.drop_constraint('uq_user_cookie')

    op.drop_column('user', 'cookie_id')





from nbconvert.preprocessors import Preprocessor
from traitlets import Bool, List, Unicode


class NbGraderPreprocessor(Preprocessor):

    default_language = Unicode("ipython")
    display_data_priority = List(
        [
            "text/html",
            "application/pdf",
            "text/latex",
            "image/svg+xml",
            "image/png",
            "image/jpeg",
            "text/plain",
        ]
    )
    enabled = Bool(
        True, help="Whether to use this preprocessor when running nbgrader"
    ).tag(config=True)



import glob
import importlib
import os
import re
import shutil
import traceback
import typing
from textwrap import dedent

import sqlalchemy
from nbconvert.exporters import Exporter, NotebookExporter
from nbconvert.exporters.exporter import ResourcesDict
from nbconvert.writers import FilesWriter
from traitlets import (
    Any,
    Bool,
    Dict,
    Instance,
    Integer,
    List,
    TraitError,
    Type,
    default,
    validate,
)
from traitlets.config import Config, LoggingConfigurable

from grader_service.convert.gradebook.gradebook import Gradebook
from grader_service.convert.nbgraderformat import SchemaTooNewError, SchemaTooOldError
from grader_service.convert.nbgraderformat.common import ValidationError
from grader_service.convert.preprocessors.execute import UnresponsiveKernelError


class GraderConvertException(Exception):
    pass


class BaseConverter(LoggingConfigurable):
    notebooks = List([])
    writer = Instance(FilesWriter)
    exporter = Instance(Exporter)
    exporter_class = Type(NotebookExporter, klass=Exporter).tag(config=True)
    preprocessors = List([])

    force = Bool(False, help="Whether to overwrite existing files").tag(config=True)

    ignore = List(
        [
            ".ipynb_checkpoints",
            "*.pyc",
            "__pycache__",
            "feedback",
            ".git",
            "grader_config.py",
        ],
        help=dedent(
            """
            List of file names or file globs.
            Upon copying directories recursively, matching files and
            directories will be ignored.
            """
        ),
    ).tag(config=True)

    pre_convert_hook = Any(
        None,
        config=True,
        allow_none=True,
        help=dedent(
            """
        An optional hook function that you can implement to do some
        bootstrapping work before converting. 
        This function is called before the notebooks are converted
        and should be used for specific converters such as Autograde,
        GenerateAssignment or GenerateFeedback.

        It will be called as (all arguments are passed as keywords)::

            hook(notebooks=notebooks, input_dir=input_dir, output_dir=output_dir)
        """
        ),
    )

    post_convert_hook = Any(
        None,
        config=True,
        allow_none=True,
        help=dedent(
            """
        An optional hook function that you can implement to do some
        work after converting. 
        This function is called after the notebooks are converted
        and should be used for specific converters such as Autograde,
        GenerateAssignment or GenerateFeedback.

        It will be called as (all arguments are passed as keywords)::

            hook(notebooks=notebooks, input_dir=input_dir, output_dir=output_dir)
        """
        ),
    )

    permissions = Integer(
        help=dedent(
            """
            Permissions to set on files output by nbgrader. The default is
            generally read-only (444), with the exception of nbgrader
            generate_assignment and nbgrader generate_feedback, in which case
            the user also has write permission.
            """
        )
    ).tag(config=True)

    @default("permissions")
    def _permissions_default(self) -> int:
        return 664

    @validate("pre_convert_hook")
    def _validate_pre_convert_hook(self, proposal):
        value = proposal["value"]
        if isinstance(value, str):
            module, function = value.rsplit(".", 1)
            value = getattr(importlib.import_module(module), function)
        if not callable(value):
            raise TraitError("pre_convert_hook must be callable")
        return value

    @validate("post_convert_hook")
    def _validate_post_convert_hook(self, proposal):
        value = proposal["value"]
        if isinstance(value, str):
            module, function = value.rsplit(".", 1)
            value = getattr(importlib.import_module(module), function)
        if not callable(value):
            raise TraitError("post_convert_hook must be callable")
        return value

    # Override this method to allow certain traitlets to be overridden
    # def _allowed_override_keys(self):
    #     return {
    #         'ClearSolutions': ['code_stub'],
    #         'ClearHiddenTests': ['begin_test_delimeter', 'end_test_delimeter'],
    #     }
    
    # Sanitize config to only allow certain keys to be overridden
    # def _sanitize_config(self, raw_config, allowed_keys):
    #     sanitized_config = Config()
    #     for key, value in raw_config.items():
    #         if key in allowed_keys:
    #             if isinstance(value, Config):
    #                 sanitized_config[key] = self._sanitize_config(value, allowed_keys[key])
    #             else:
    #                 sanitized_config[key] = value
    #     return sanitized_config

    def __init__(
            self, input_dir: str, output_dir: str, file_pattern: str, copy_files: bool, **kwargs: typing.Any
    ) -> None:
        super(BaseConverter, self).__init__(**kwargs)
        self._input_directory = os.path.abspath(os.path.expanduser(input_dir))
        self._output_directory = os.path.abspath(os.path.expanduser(output_dir))
        self._file_pattern = file_pattern
        self._copy_files = copy_files
        if self.parent and hasattr(self.parent, "logfile"):
            self.logfile = self.parent.logfile
        else:
            self.logfile = None

        c = Config()

        custom_config_path = f'{self._input_directory}/grader_config.py'
        if os.path.exists(custom_config_path):

            local_vars = {'c': c}
            with open(custom_config_path) as f:
                code = compile(f.read(), 'config.py', 'exec')
                exec(code, {}, local_vars)

            c = local_vars['c']

        c.Exporter.default_preprocessors = []
        self.update_config(c)

    # register pre-processors to self.exporter
    # self.convert_notebooks() converts all notebooks in the CourseDir
    # notebooks are set in init_notebooks()
    def start(self) -> None:
        self.init_notebooks()
        self.writer = FilesWriter(parent=self, config=self.config)
        self.exporter: Exporter = self.exporter_class(parent=self, config=self.config)
        for pp in self.preprocessors:
            self.exporter.register_preprocessor(pp)
        currdir = os.getcwd()
        os.chdir(self._output_directory)
        try:
            self.convert_notebooks()
        finally:
            os.chdir(currdir)

    @default("classes")
    def _classes_default(self):
        classes = super(BaseConverter, self)._classes_default()
        classes.append(FilesWriter)
        classes.append(Exporter)
        for pp in self.preprocessors:
            if len(pp.class_traits(config=True)) > 0:
                classes.append(pp)
        return classes

    # returns string that can be used for globs
    def _format_source(self, escape: bool = False) -> str:
        source = os.path.join(self._input_directory, self._file_pattern)
        if escape:
            return re.escape(source)
        else:
            return source

    def init_notebooks(self) -> None:
        self.notebooks = []
        notebook_glob = self._format_source()
        self.notebooks = glob.glob(notebook_glob)
        if len(self.notebooks) == 0:
            self.log.warning("No notebooks were matched by '%s'", notebook_glob)

    def init_single_notebook_resources(
            self, notebook_filename: str
    ) -> typing.Dict[str, typing.Any]:
        resources = {}
        resources["unique_key"] = os.path.splitext(os.path.basename(notebook_filename))[
            0
        ]
        resources["output_files_dir"] = "%s_files" % os.path.basename(notebook_filename)
        resources["output_json_file"] = "gradebook.json"
        resources["output_json_path"] = os.path.join(
            self._output_directory, resources["output_json_file"]
        )
        resources["nbgrader"] = dict()  # support nbgrader pre-processors
        return resources

    def write_single_notebook(self, output: str, resources: ResourcesDict) -> None:
        # configure the writer build directory
        self.writer.build_directory = self._output_directory

        # write out the results
        self.writer.write(output, resources, notebook_name=resources["unique_key"])

    def init_destination(self) -> bool:
        """Initialize the destination for an assignment. Returns whether the
        assignment should actually be processed or not (i.e. whether the
        initialization was successful).

        """
        dest = self._output_directory
        source = self._input_directory

        # if we have specified --force, then always remove existing stuff
        if self.force:
            return True

        # if files exist in the destination and force is not specified return false
        src_files = glob.glob(self._format_source())
        for src in src_files:
            file_name = os.path.join(dest, os.path.relpath(src, source))
            if os.path.exists(os.path.join(dest, file_name)):
                return False
        return True

    def copy_unmatched_files(self, gb: Gradebook):
        """
        Copy the files from source to the output directory that were not matched by the file pattern.
        :return: None
        """
        dst = self._output_directory
        src = self._input_directory

        if self._copy_files:
            copied_files = []

            def save_copy2(src_c, dst_c, *, follow_symlinks=True):
                shutil.copy2(src_c, dst_c, follow_symlinks=follow_symlinks)
                rel_file_name = os.path.relpath(src_c, src)
                copied_files.append(rel_file_name)

            self.log.info(f"Copying unmatched files from {src} to {dst}")
            ignore = shutil.ignore_patterns(*self.ignore + [self._file_pattern])
            shutil.copytree(src, dst, copy_function=save_copy2, ignore=ignore, dirs_exist_ok=True)
            gb.set_extra_files(copied_files)
        else:
            for rel_path in gb.get_extra_files():
                src_file = os.path.join(src, rel_path)
                if not os.path.isfile(src_file):
                    self.log.warning(f"The file {rel_path} cannot be copied because it does not exist in {src}!")
                    continue

                # make sure the subdirectories of the file exists before copying
                rel_dir = os.path.dirname(rel_path)
                os.makedirs(os.path.join(dst, rel_dir), exist_ok=True)

                dst_file = os.path.join(dst, rel_path)
                shutil.copy2(src_file, dst_file)

    def set_permissions(self) -> None:
        self.log.info("Setting destination file permissions to %s", self.permissions)
        dest = os.path.normpath(self._output_directory)
        permissions = int(str(self.permissions), 8)
        for dirname, _, filenames in os.walk(dest):
            for filename in filenames:
                os.chmod(os.path.join(dirname, filename), permissions)

    def convert_single_notebook(self, notebook_filename: str) -> None:
        """
        Convert a single notebook.

        Performs the following steps:
            1. Initialize notebook resources
            2. Export the notebook to a particular format
            3. Write the exported notebook to file
        """
        self.log.info("Converting notebook %s", notebook_filename)
        resources = self.init_single_notebook_resources(notebook_filename)
        output, resources = self.exporter.from_filename(
            notebook_filename, resources=resources
        )
        self.write_single_notebook(output, resources)

    def convert_notebooks(self) -> None:
        errors = []

        def _handle_failure(exception) -> None:
            # dest = os.path.normpath(self._output_directory)
            # rmtree(dest)
            pass

        # initialize the list of notebooks and the exporter
        self.notebooks = sorted(self.notebooks)

        try:
            # determine whether we actually even want to process the notebooks
            should_process = self.init_destination()
            if not should_process:
                return

            self.run_pre_convert_hook()

            # convert all the notebooks
            for notebook_filename in self.notebooks:
                self.convert_single_notebook(notebook_filename)

            # set assignment permissions
            self.set_permissions()
            self.run_post_convert_hook()

            json_path = os.path.join(self._output_directory, "gradebook.json")
            with Gradebook(json_path) as gb:
                self.copy_unmatched_files(gb)

        except UnresponsiveKernelError as e:
            self.log.error(
                "While processing files %s, the kernel became "
                "unresponsive and we could not interrupt it. This probably "
                "means that the students' code has an infinite loop that "
                "consumes a lot of memory or something similar. nbgrader "
                "doesn't know how to deal with this problem, so you will "
                "have to manually edit the students' code (for example, to "
                "just throw an error rather than enter an infinite loop). ",
                self._format_source(),
            )
            errors.append(e)
            _handle_failure(e)

        except SchemaTooOldError as e:
            _handle_failure(e)
            msg = (
                "One or more notebooks in the assignment use an old version \n"
                "of the nbgrader metadata format. Please **back up your class files \n"
                "directory** and then update the metadata using:\n\nnbgrader update .\n"
            )
            self.log.error(msg)
            raise GraderConvertException(msg)

        except SchemaTooNewError as e:
            _handle_failure(e)
            msg = (
                "One or more notebooks in the assignment use an newer version \n"
                "of the nbgrader metadata format. Please update your version of \n"
                "nbgrader to the latest version to be able to use this notebook.\n"
            )
            self.log.error(msg)
            raise GraderConvertException(msg)

        except KeyboardInterrupt as e:
            _handle_failure(e)
            self.log.error("Canceled")
            raise

        except ValidationError as e:
            _handle_failure(e)
            self.log.error(e.message)
            raise GraderConvertException(e.message)

        except Exception as e:
            self.log.error(
                "There was an error processing files: %s", self._format_source()
            )
            self.log.error(traceback.format_exc())
            errors.append(e)
            _handle_failure(e)

        if len(errors) > 0:
            if self.logfile:
                msg = (
                    "Please see the error log ({}) for details on the specific "
                    "errors on the above failures.".format(self.logfile)
                )
            else:
                msg = errors[0]

            self.log.error(msg)
            raise GraderConvertException(msg)

    def run_pre_convert_hook(self):
        if self.pre_convert_hook:
            self.log.info("Running pre-convert hook")
            try:
                self.pre_convert_hook(
                    notebooks=self.notebooks,
                    input_dir=self._input_directory,
                    output_dir=self._output_directory,
                )
            except Exception:
                self.log.error("Pre-convert hook failed", exc_info=True)

    def run_post_convert_hook(self):
        if self.post_convert_hook:
            self.log.info("Running post-convert hook")
            try:
                self.post_convert_hook(
                    notebooks=self.notebooks,
                    input_dir=self._input_directory,
                    output_dir=self._output_directory,
                )
            except Exception:
                self.log.error("Post-convert hook failed", exc_info=True)



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import enum
from datetime import datetime

from grader_service.api.models import lecture
from sqlalchemy import Column, Enum, Integer, String, DateTime
from sqlalchemy.orm import relationship

from grader_service.orm.base import Base, DeleteState, Serializable


class LectureState(enum.IntEnum):
    active = 0
    complete = 1


class Lecture(Base, Serializable):
    __tablename__ = "lecture"
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(255), nullable=True, unique=False)
    code = Column(String(255), nullable=True, unique=True)
    state = Column(Enum(LectureState), nullable=False, unique=False)
    deleted = Column(Enum(DeleteState), nullable=False, unique=False)

    created_at = Column(DateTime, default=datetime.utcnow, nullable=False)
    updated_at = Column(DateTime, default=datetime.utcnow,
                        onupdate=datetime.utcnow, nullable=False)

    assignments = relationship("Assignment", back_populates="lecture")
    roles = relationship("Role", back_populates="lecture")
    groups = relationship("Group", back_populates="lecture")

    @property
    def model(self) -> lecture.Lecture:
        return lecture.Lecture(
            id=self.id,
            name=self.name,
            code=self.code,
            complete=self.state == LectureState.complete,
        )



import os
import argparse
from grader_service.autograding.celery.app import CeleryApp


def main():
    parser = argparse.ArgumentParser(prog='Grader Worker', description='Starts celery worker for grader service.')
    parser.add_argument('-f', '--config', help='config file path', required=True)

    args = parser.parse_args()

    celery = CeleryApp.instance(config_file=os.path.abspath(args.config))
    app = celery.app

    worker = app.Worker(**celery.worker_kwargs)
    worker.start()


if __name__ == '__main__':
    main()



# Copyright (c) 2022, TU Wien
# All rights reserved.
# grader service orm
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
from datetime import datetime

from grader_service.api.models import submission
from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    Enum,
    ForeignKey,
    Float,
    Integer,
    String,
)
from sqlalchemy.orm import relationship

from grader_service.orm.base import Base, DeleteState, Serializable


class Submission(Base, Serializable):
    __tablename__ = "submission"
    id = Column(Integer, primary_key=True, autoincrement=True)
    date = Column(DateTime, nullable=False)
    auto_status = Column(
        Enum("pending", "not_graded",
             "automatically_graded", "grading_failed"),
        default="not_graded",
        nullable=False,
    )
    manual_status = Column(Enum("not_graded", "manually_graded",
                                "being_edited"))
    score = Column(Float, nullable=True)
    assignid = Column(Integer, ForeignKey("assignment.id"))
    username = Column(String(255), ForeignKey("user.name"))
    commit_hash = Column(String(length=40), nullable=False)
    feedback_status = Column(
        Enum("not_generated", "generating", "generated", "generation_failed", "feedback_outdated"),
        default="not_generated",
        nullable=False)
    deleted = Column(Enum(DeleteState), nullable=False, unique=False, default=DeleteState.active)
    edited = Column(Boolean, nullable=False)
    updated_at = Column(DateTime, default=datetime.utcnow,
                        onupdate=datetime.utcnow, nullable=False)
    grading_score = Column(Float, nullable=False)
    score_scaling = Column(Float, server_default="1.0", nullable=False)

    assignment = relationship("Assignment", back_populates="submissions")
    user = relationship("User", back_populates="submissions")
    logs = relationship("SubmissionLogs",
                        back_populates="submission", uselist=False)
    properties = relationship("SubmissionProperties",
                              back_populates="submission", uselist=False)

    @property
    def model(self) -> submission.Submission:
        return submission.Submission(
            id=self.id,
            submitted_at=None
            if self.date is None
            else (self.date.isoformat("T", "milliseconds") ),
            username=self.username,
            auto_status=self.auto_status,
            manual_status=self.manual_status,
            score_scaling=self.score_scaling,
            grading_score=self.grading_score,
            score=self.score,
            assignid=self.assignid,
            commit_hash=self.commit_hash,
            feedback_status=self.feedback_status,
            edited=self.edited
        )



import logging
from typing import Any, Dict, List

from grader_service.auth.login import LogoutHandler
from grader_service.auth.oauth2 import OAuthLogoutHandler
from grader_service.server import GRADER_COOKIE_NAME
from grader_service.utils import url_path_join  # type: ignore
from traitlets import CaselessStrEnum
from traitlets import List as TraitletsList
from traitlets import Set as TraitletsSet
from traitlets import Unicode

from grader_service.utils import get_browser_protocol
from .constants import LTI13_CUSTOM_CLAIM
from .error import LoginError
from .handlers import LTI13CallbackHandler, LTI13ConfigHandler, LTI13LoginInitHandler
from ..auth import Authenticator

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class LTI13LogoutHandler(LogoutHandler):
    async def render_logout_page(self):
        """Render the logout page, if any

        Override this function to set a custom logout page.
        """
        html = await self.render_template('auth/logout.html.j2')
        self.finish(html)


class LTI13Authenticator(Authenticator):
    """
    JupyterHub LTI 1.3 Authenticator. LTI 1.3 is basically an extension of OIDC/OAuth2.
    Messages sent to this authenticator are sent from a LTI 1.3 Platform, such as an LMS.
    JupyterHub, as the authenticator, works as the LTI 1.3 External Tool.
    The basic login flow is authentication using the implicit flow.
    As such, at least one client id is required.

    Ref:
      - http://www.imsglobal.org/spec/lti/v1p3/
      - https://openid.net/specs/openid-connect-core-1_0.html#ImplicitFlowAuth
    """

    login_service = "LTI 1.3"

    # handlers used for login, callback, and json config endpoints
    login_handler = LTI13LoginInitHandler
    callback_handler = LTI13CallbackHandler
    config_handler = LTI13ConfigHandler
    logout_handler = LTI13LogoutHandler

    authorize_url = Unicode(
        config=True,
        help="""Authorization end-point of the platforms identity provider.""",
    )

    client_id = TraitletsSet(
        trait=Unicode(),
        config=True,
        help="""
        The client ID or a list of client IDs identifying the JuyterHub within the LMS platform.
        Must contain the client IDs created when registering the tool on the LMS platform.

        Possible values are of type str or iterables thereof.
        """,
    )

    jwks_algorithms = TraitletsList(
        default_value=["RS256"],
        config=True,
        help="""
        Supported algorithms for signing JWT. The actual algorithm is declared by the authenticator
        in the `id_token_signed_response_alg` parameter during out-of-band registration.

        References:
        https://www.imsglobal.org/spec/security/v1p0/#authentication-response-validation
        """,
    )

    issuer = Unicode(
        allow_none=False,
        config=True,
        help="""
        The platform's issuer identifier. It is a case-sensitive URL, using the HTTPS
        scheme, that contains scheme, host, and optionally, port number, and path components,
        and no query or fragment components. It is provided by the platform.
        """,
    )

    jwks_endpoint = Unicode(
        allow_none=False,
        config=True,
        help="""
        The platform's JWKS endpoint used to obtain it's JSON Web Key Set to validate JWT signatures.
        """,
    )

    username_key = Unicode(
        "email",
        allow_none=False,
        config=True,
        help="""
        JWT claim present in LTI 1.3 login initiation flow used to set the user's JupyterHub's username.
        Some common examples include:

          - User's email address: email
          - Given name: given_name
        
        Your LMS (Canvas / Open EdX / Moodle / others) may provide additional keys in the
        LTI 1.3 login initiation flow that you can use to set the username. In most cases these
        are located in the `https://purl.imsglobal.org/spec/lti/claim/custom` claim. In this case,
        `username_key` must be prefixed with "custom_". For example, `username_key` value "custom_uname"
        will set the username to the value of the parameter `uname` within the
        `https://purl.imsglobal.org/spec/lti/claim/custom` claim.
        
        If your platform's LTI 1.3 settings are defined with privacy enabled, then by default the `sub`
        claim is used to set the username.
        
        You may also have the option of using variable substitutions to fetch values that aren't provided with
        your vendor's standard LTI 1.3 login initiation flow request.

        Reference to the IMS LTI specification on variable substitutions:
        http://www.imsglobal.org/spec/lti/v1p3/#customproperty.
        """,
    )

    uri_scheme = CaselessStrEnum(
        ("auto", "https", "http"),
        default_value="auto",
        config=True,
        help="""
        Scheme to use for endpoint URLs offered by this authenticator.

        Possible values are "auto", "https" and "http", where "auto" is the default.
        When "auto" is chosen, the scheme is inferred from the incomming request's header.
        Since this may lead to unreliable results in some deployment scenarios (in particular
        when several different versions of forwarding headers are mixed), manually specifying it
        here is kept as an escape hatch.
        """,
    )

    tool_name = Unicode(
        "JupyterHub",
        config=True,
        help="""
        Name of tool provided to the LMS when installed via the config URL.

        This is primarily used for display purposes.
        """,
    )

    tool_description = Unicode(
        "Launch interactive Jupyter Notebooks with JupyterHub",
        config=True,
        help="""
        Description of tool provided to the LMS when installed via the config URL.

        This is primarily used for display purposes.
        """,
    )

    def login_url(self, base_url):
        return url_path_join(base_url, "lti13", "oauth_login")

    def callback_url(self, base_url):
        return url_path_join(base_url, "lti13", "oauth_callback")

    def config_json_url(self, base_url):
        return url_path_join(base_url, "lti13", "config")

    def get_handlers(self, base_url_path):
        self.log.info(base_url_path)
        return [
            (self.login_url(base_url_path), self.login_handler),
            (self.callback_url(base_url_path), self.callback_handler),
            (self.config_json_url(base_url_path), self.config_handler),
            (self.logout_url(base_url_path), self.logout_handler),
        ]

    async def authenticate(
            self, handler: LTI13LoginInitHandler, data: Dict[str, str] = None
    ) -> Dict[str, Any]:
        """
        Handles LTI 1.3 launch requests based on a passed JWT.

        Args:
          handler: handler object
          data: authentication dictionary. The decoded, verified and validated id_token send by the platform

        Returns:
          Authentication dictionary
        """
        if not data:
            data = {}

        username = self.get_username(data)

        return {
            "name": username,
            "auth_state": data,
        }

    def get_username(self, token: Dict[str, Any]) -> str:
        """
        Infer the username from the ID token.

        If username_key begins with "custom_", that part will be removed and the key will
        be looked up inside the custom claim of the ID token.
        """
        data = token
        username_key = self.username_key

        if not username_key:
            username_key = "sub"

        if username_key.startswith("custom_"):
            data = token.get(LTI13_CUSTOM_CLAIM, {})
            # when dropping support for Python 3.8 we can replace the following by `username_key.removeprefix`
            username_key = username_key[len("custom_"):]

        username = data.get(username_key)
        if not username:
            logger.warning(
                f"Cannot find the key {username_key} in the ID token. `sub` used instread."
            )
            username = token.get("sub")
        if not username:
            raise LoginError(
                f"Unable to set the username with username_key {username_key}"
            )
        return username

    def get_uri_scheme(self, request) -> str:
        """Return scheme to use for endpoint URLs of this authenticator."""
        if self.uri_scheme == "auto":
            return get_browser_protocol(request)
        # manually specified https or http
        return self.uri_scheme



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import io
import json
import logging
import os
import shlex
import shutil
import stat
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from subprocess import Popen, PIPE, CalledProcessError
from typing import Optional

from traitlets.config import Config

from grader_service.convert.converters.autograde import Autograde
from grader_service.convert.gradebook.models import GradeBookModel
from grader_service.orm.assignment import Assignment
from grader_service.orm.group import Group
from grader_service.orm.lecture import Lecture
from grader_service.orm.submission import Submission
from sqlalchemy.orm import Session
from traitlets.config.configurable import LoggingConfigurable

from traitlets.traitlets import TraitError, Unicode, validate, Callable

from grader_service.orm.submission_logs import SubmissionLogs
from grader_service.orm.submission_properties import SubmissionProperties


def rm_error(func, path, exc_info):
    if not os.access(path, os.W_OK):
        # Is the error an access error ?
        os.chmod(path, stat.S_IWUSR)
        func(path)
    else:
        raise


@dataclass
class AutogradingStatus:
    status: str
    started_at: datetime
    finished_at: datetime


def default_timeout_func(l: Lecture) -> int:
    return 360


class LocalAutogradeExecutor(LoggingConfigurable):
    """
    Runs an autograde job on the local machine
    with the current Python environment.
    Sets up the necessary directories
    and the gradebook JSON file used by :mod:`grader_service.convert`.
    """
    relative_input_path = Unicode("convert_in",
                                  allow_none=True).tag(config=True)
    relative_output_path = Unicode("convert_out",
                                   allow_none=True).tag(config=True)
    git_executable = Unicode("git", allow_none=False).tag(config=True)

    timeout_func = Callable(default_timeout_func, allow_none=False,
                            help="Function that takes a lecture as an argument and returns the cell timeout in seconds."
                            ).tag(config=True)

    def __init__(self, grader_service_dir: str,
                 submission: Submission, close_session=True, **kwargs):
        """
        Creates the executor in the input
        and output directories that are specified
        by :attr:`base_input_path` and :attr:`base_output_path`.
        The grader service directory is used for accessing
        the git repositories to push the grading results.
        The database session is retrieved from the submission object.
        The associated session of the submission has to be available
        and must not be closed beforehand.

        :param grader_service_dir: The base directory of the whole
        grader service specified in the configuration.
        :type grader_service_dir: str
        :param submission: The submission object
        which should be graded by the executor.
        :type submission: Submission
        """
        super(LocalAutogradeExecutor, self).__init__(**kwargs)
        self.grader_service_dir = grader_service_dir
        self.submission = submission
        self.assignment: Assignment = submission.assignment
        self.session: Session = Session.object_session(self.submission)
        # close session after grading (might need session later)
        self.close_session = close_session

        self.autograding_start: Optional[datetime] = None
        self.autograding_finished: Optional[datetime] = None
        self.autograding_status: Optional[str] = None
        self.grading_logs: Optional[str] = None

    def start(self):
        """
        Starts the autograding job.
        This is the only method that is exposed to the client.
        It re-raises all exceptions that happen while running.
        """
        self.log.info(f"Starting autograding job for submission "
                      f"{self.submission.id} in {self.__class__.__name__}")
        try:
            self._pull_submission()
            self.autograding_start = datetime.now()
            self._run()
            self.autograding_finished = datetime.now()
            self._set_properties()
            self._push_results()
            self._set_db_state()
            ts = round((self.autograding_finished - self.autograding_start)
                       .total_seconds())
            self.log.info(
                f"Successfully completed autograding job for submission "
                f"{self.submission.id} in {self.__class__.__name__};"
                + f" took {ts // 60}min {ts % 60}s")
        except Exception:
            self.log.error(
                f"Failed autograding job for submission "
                f"{self.submission.id} in {self.__class__.__name__}",
                exc_info=True)
            self._set_db_state(success=False)
        finally:
            self._cleanup()

    @property
    def input_path(self):
        return os.path.join(self.grader_service_dir, self.relative_input_path,
                            f"submission_{self.submission.id}")

    @property
    def output_path(self):
        return os.path.join(self.grader_service_dir, self.relative_output_path,
                            f"submission_{self.submission.id}")

    def _write_gradebook(self, gradebook_str: str):
        """
        Writes the gradebook to the output directory where it will be used by
        :mod:`grader_service.convert` to load the data.
        The name of the written file is gradebook.json.
        :param gradebook_str: The content of the gradebook.
        :return: None
        """
        if not os.path.exists(self.output_path):
            os.mkdir(self.output_path)
        path = os.path.join(self.output_path, "gradebook.json")
        self.log.info(f"Writing gradebook to {path}")
        with open(path, "w") as f:
            f.write(gradebook_str)

    def _pull_submission(self):
        """
        Pulls the submission repository into the input path
        based on the assignment type.
        :return: Coroutine
        """
        if not os.path.exists(self.input_path):
            Path(self.input_path).mkdir(parents=True, exist_ok=True)

        assignment: Assignment = self.submission.assignment
        lecture: Lecture = assignment.lecture

        if assignment.type == "user":
            repo_name = self.submission.username
        else:
            # TODO: fix query to work with group.name
            group = self.session.query(Group).get(
                (self.submission.username, lecture.id)
            )
            if group is None:
                raise ValueError()
            repo_name = group.name

        if self.submission.edited:
            git_repo_path = os.path.join(
                self.grader_service_dir,
                "git",
                lecture.code,
                str(assignment.id),
                "edit",
                str(self.submission.id)
            )
        else:
            git_repo_path = os.path.join(
                self.grader_service_dir,
                "git",
                lecture.code,
                str(assignment.id),
                assignment.type,
                repo_name,
            )

        if os.path.exists(self.input_path):
            shutil.rmtree(self.input_path, onerror=rm_error)
        os.mkdir(self.input_path)

        self.log.info(f"Pulling repo {git_repo_path} into input directory")

        command = f"{self.git_executable} init"
        self.log.info(f"Running {command}")
        self._run_subprocess(command, self.input_path)

        command = f'{self.git_executable} pull "{git_repo_path}" main'
        self.log.info(f"Running {command}")
        self._run_subprocess(command, self.input_path)
        self.log.info("Successfully cloned repo")

        # Checkout to commit of submission except when it was manually edited
        if not self.submission.edited:
            command = f"{self.git_executable} checkout " \
                      f"{self.submission.commit_hash}"
            self.log.info(f"Running {command}")
            self._run_subprocess(command, self.input_path)
            self.log.info(f"Now at commit {self.submission.commit_hash}")

    def _run(self):
        """
        Runs the autograding in the current interpreter
        and captures the output.
        :return: Coroutine
        """
        if os.path.exists(self.output_path):
            shutil.rmtree(self.output_path, onerror=rm_error)

        os.makedirs(self.output_path, exist_ok=True)
        self._write_gradebook(self._put_grades_in_assignment_properties())

        c = Config()
        c.ExecutePreprocessor.timeout = self.timeout_func(self.assignment.lecture)

        autograder = Autograde(self.input_path, self.output_path, "*.ipynb",
                               copy_files=self.assignment.allow_files, config=c)
        autograder.force = True

        log_stream = io.StringIO()
        log_handler = logging.StreamHandler(log_stream)
        autograder.log.addHandler(log_handler)

        try:
            autograder.start()
        finally:
            self.grading_logs = log_stream.getvalue()
            autograder.log.removeHandler(log_handler)

    def _put_grades_in_assignment_properties(self) -> str:
        """
        Checks if assignment was already graded and returns updated properties.
        :return: str
        """
        if self.submission.manual_status == "not_graded":
            return self.assignment.properties

        assignment_properties = json.loads(self.assignment.properties)
        submission_properties = json.loads(
            self.submission.properties.properties)
        notebooks = set.intersection(
            set(assignment_properties["notebooks"].keys()),
            set(submission_properties["notebooks"].keys()))
        for notebook in notebooks:
            # Set grades
            #
            assignment_properties["notebooks"][notebook]["grades_dict"] = \
                submission_properties["notebooks"][notebook]["grades_dict"]
            # Set comments
            assignment_properties["notebooks"][notebook]["comments_dict"] = \
                submission_properties["notebooks"][notebook]["comments_dict"]

        properties_str = json.dumps(assignment_properties)
        self.log.info("Added grades dict to properties")
        return properties_str

    def _push_results(self):
        """
        Pushes the results to the autograde repository
        as a separate branch named after the commit hash of the submission.
        Removes the gradebook.json file before doing so.
        :return: Coroutine
        """
        os.unlink(os.path.join(self.output_path, "gradebook.json"))
        self.log.info(f"Pushing files: {os.listdir(self.output_path)}")

        assignment: Assignment = self.submission.assignment
        lecture: Lecture = assignment.lecture

        if assignment.type == "user":
            repo_name = self.submission.username
        else:
            group = self.session.query(Group).get(
                (self.submission.username, lecture.id)
            )
            if group is None:
                raise ValueError()
            repo_name = group.name

        git_repo_path = os.path.join(
            self.grader_service_dir,
            "git",
            lecture.code,
            str(assignment.id),
            "autograde",
            assignment.type,
            repo_name,
        )

        if not os.path.exists(git_repo_path):
            os.makedirs(git_repo_path, exist_ok=True)
            try:
                self._run_subprocess(
                    f'git init --bare "{git_repo_path}"', self.output_path
                )
            except CalledProcessError:
                raise

        command = f"{self.git_executable} init"
        self.log.info(f"Running {command} at {self.output_path}")
        try:
            self._run_subprocess(command, self.output_path)
        except CalledProcessError:
            pass

        self.log.info(f"Creating new branch "
                      f"submission_{self.submission.commit_hash}")
        command = (
            f"{self.git_executable} switch -c "
            f"submission_{self.submission.commit_hash}"
        )
        try:
            self._run_subprocess(command, self.output_path)
        except CalledProcessError:
            pass
        self.log.info(f"Now at branch "
                      f"submission_{self.submission.commit_hash}")

        self.log.info(f"Commiting all files in {self.output_path}")
        try:
            self._run_subprocess(
                f"{self.git_executable} add -A", self.output_path
            )
            self._run_subprocess(
                f'{self.git_executable} commit -m '
                f'"{self.submission.commit_hash}"',
                self.output_path,
            )
        except CalledProcessError:
            raise RuntimeError("Failed to commit changes")

        self.log.info(
            f"Pushing to {git_repo_path} at branch "
            f"submission_{self.submission.commit_hash}"
        )
        command = f'{self.git_executable} push -uf ' \
                  f'"{git_repo_path}" submission_{self.submission.commit_hash}'
        try:
            self._run_subprocess(command, self.output_path)
        except CalledProcessError:
            raise RuntimeError(f"Failed to push to {git_repo_path}")
        self.log.info("Pushing complete")

    def _set_properties(self):
        """
        Loads the contents of the gradebook.json file
        and sets them as the submission properties.
        Also calculates the score of the submission
        after autograding based on the updated properties.
        :return: None
        """
        with open(os.path.join(self.output_path, "gradebook.json"), "r") as f:
            gradebook_str = f.read()

        properties = SubmissionProperties(properties=gradebook_str,
                                          sub_id=self.submission.id)

        self.session.merge(properties)

        gradebook_dict = json.loads(gradebook_str)
        book = GradeBookModel.from_dict(gradebook_dict)
        score = 0
        for id, n in book.notebooks.items():
            score += n.score
        self.submission.grading_score = score
        self.submission.score = self.submission.score_scaling * score
        self.session.commit()

    def _set_db_state(self, success=True):
        """
        Sets the submission autograding status based on the success parameter
        and sets the logs from autograding.
        :param success: Whether the grading process was a success or failure.
        :return: None
        """
        if success:
            self.submission.auto_status = "automatically_graded"
        else:
            self.submission.auto_status = "grading_failed"

        if self.grading_logs is not None:
            self.grading_logs = self.grading_logs.replace("\x00", "")
        logs = SubmissionLogs(logs=self.grading_logs,
                              sub_id=self.submission.id)
        self.session.merge(logs)
        self.session.commit()

    def _cleanup(self):
        """
        Removes all files from the input and output directories
        and closes the session if specified by self.close_session.
        :return: None
        """
        try:
            shutil.rmtree(self.input_path)
            shutil.rmtree(self.output_path)
        except FileNotFoundError:
            pass
        if self.close_session:
            self.session.close()

    def _run_subprocess(self, command: str, cwd: str) -> Popen[bytes]:
        """
        Execute the command as a subprocess.
        :param command: The command to execute as a string.
        :param cwd: The working directory the subprocess should run in.
        :return: Coroutine which resolves to a Subprocess object
        which resulted from the execution.
        """
        try:
            process = Popen(shlex.split(command), stdout=PIPE, stderr=PIPE, cwd=cwd)
            process.wait()
        except CalledProcessError:
            self.grading_logs = process.stderr.read().decode("utf-8")
            self.log.error(self.grading_logs)
        except FileNotFoundError as e:
            self.grading_logs = str(e)
            self.log.error(self.grading_logs)
            raise e
        return process

    @validate("relative_input_path", "relative_output_path")
    def _validate_service_dir(self, proposal):
        path: str = proposal["value"]
        if not os.path.exists(self.grader_service_dir + "/" + path):
            self.log.info(f"Path {path} not found, creating new directories.")
            Path(path).mkdir(parents=True, exist_ok=True, mode=0o700)
        if not os.path.isdir(self.grader_service_dir + "/" + path):
            raise TraitError("The path has to be an existing directory")
        return path

    @validate("convert_executable", "git_executable")
    def _validate_executable(self, proposal):
        exec: str = proposal["value"]
        if shutil.which(exec) is None:
            raise TraitError(f"The executable is not valid: {exec}")
        return exec


class LocalProcessAutogradeExecutor(LocalAutogradeExecutor):
    """Runs an autograde job on the local machine
    with the default Python environment in a separate process.
    Sets up the necessary directories
    and the gradebook JSON file used by :mod:`grader_service.convert`.
    """

    convert_executable = Unicode("grader-convert",
                                 allow_none=False).tag(config=True)

    def _run(self):
        """
        Runs the autograding in a separate python interpreter
        as a sub-process and captures the output.

        :return: Coroutine
        """
        if os.path.exists(self.output_path):
            shutil.rmtree(self.output_path, onerror=rm_error)

        os.mkdir(self.output_path)
        self._write_gradebook(self._put_grades_in_assignment_properties())

        command = f'{self.convert_executable} autograde ' \
                  f'-i "{self.input_path}" ' \
                  f'-o "{self.output_path}" ' \
                  f'-p "*.ipynb" ' \
                  f'--copy_files={self.assignment.allow_files} ' \
                  f'--ExecutePreprocessor.timeout={self.timeout_func(self.assignment.lecture)}'
        self.log.info(f"Running {command}")
        process = self._run_subprocess(command, None)
        if process.returncode == 0:
            self.grading_logs = process.stderr.read().decode("utf-8")
            self.log.info(self.grading_logs)
            self.log.info("Process has successfully completed execution!")
        else:
            raise RuntimeError("Process has failed execution!")





# coding: utf-8

import os
import traceback

from nbformat import current_nbformat
from nbformat import read as orig_read
from nbformat import write as orig_write
from traitlets import Bool

from grader_service.convert.nbgraderformat import (
    MetadataValidator,
    SchemaTooNewError,
    ValidationError,
    write,
)
from grader_service.convert.utils import find_all_notebooks
from grader_service.convert.converters.baseapp import ConverterApp

aliases = {
    "log-level": "Application.log_level",
}
flags = {}


class UpdateApp(ConverterApp):

    name = u"update"
    description = u"Update nbgrader notebook metadata"

    aliases = aliases
    flags = flags

    validate = Bool(True, help="whether to validate metadata after updating it").tag(
        config=True
    )

    examples = """
        nbgrader stores metadata in the JSON source of the notebooks. Previously,
        we did not do a good job of validating whether this metadata was
        correctly formatted or not. Starting in version 0.4.0 of nbgrader, we
        are explicitly validating this metadata. This will require that you
        update the metadata in your old nbgrader notebooks to be consistent
        with what nbgrader expects.

        The `nbgrader update` command performs this metadata update for you
        easily. All you need to do is point it at a directory, and it will
        find all notebooks in that directory and update them to have the
        correct metadata:

            # update notebooks rooted in the current directory
            nbgrader update .

            # update notebooks rooted in the `class_files` directory
            nbgrader update class_files/

        Alternately, you can open all your notebooks with the "Create Assignment"
        toolbar and re-save them from the notebook interface. But, it will be
        more efficient to run the `nbgrader update` command to get them all in
        one fell swoop.
        """

    def start(self):
        super(UpdateApp, self).start()

        if len(self.extra_args) < 1:
            self.fail(
                "No notebooks or directories given. Usage:\n\n"
                "nbgrader update <NOTEBOOK>\n"
                "nbgrader update <DIRECTORY>\n"
            )

        notebooks = set()
        for name in self.extra_args:
            if not os.path.exists(name):
                self.fail("No such file or directory: {}".format(name))
            elif os.path.isdir(name):
                notebooks.update(
                    [os.path.join(name, x) for x in find_all_notebooks(name)]
                )
            elif not name.endswith(".ipynb"):
                self.log.warning("{} is not a notebook, ignoring".format(name))
            else:
                notebooks.add(name)

        notebooks = sorted(list(notebooks))
        for notebook in notebooks:
            self.log.info("Updating metadata for notebook: {}".format(notebook))
            nb = orig_read(notebook, current_nbformat)
            nb = MetadataValidator().upgrade_notebook_metadata(nb)
            if self.validate:
                try:
                    write(nb, notebook)
                except ValidationError:
                    self.log.error(traceback.format_exc())
                    self.fail(
                        "Notebook '{}' failed to validate, metadata is corrupted".format(
                            notebook
                        )
                    )
                except SchemaTooNewError:
                    self.log.error(traceback.format_exc())
                    self.fail(
                        (
                            "The notebook '{}' uses a newer version "
                            "of the nbgrader metadata format. Please update your version of "
                            "nbgrader to the latest version to be able to use this notebook."
                        ).format(notebook)
                    )
            else:
                orig_write(nb, notebook)





from typing import Tuple

from nbconvert.exporters.exporter import ResourcesDict
from nbformat.notebooknode import NotebookNode

from grader_service.convert import utils
from grader_service.convert.preprocessors.base import NbGraderPreprocessor


class DeduplicateIds(NbGraderPreprocessor):
    """A preprocessor to overwrite information about grade and solution cells."""

    def preprocess(
        self, nb: NotebookNode, resources: ResourcesDict
    ) -> Tuple[NotebookNode, ResourcesDict]:
        # keep track of grade ids encountered so far
        self.grade_ids = set([])

        # process each cell in reverse order
        nb, resources = super(DeduplicateIds, self).preprocess(nb, resources)


        return nb, resources

    def preprocess_cell(
        self, cell: NotebookNode, resources: ResourcesDict, cell_index: int
    ) -> Tuple[NotebookNode, ResourcesDict]:

        if not (
            utils.is_grade(cell) or utils.is_solution(cell) or utils.is_locked(cell)
        ):
            self.log.warning("cell above")
            return cell, resources

        grade_id = cell.metadata.nbgrader["grade_id"]

        if grade_id in self.grade_ids:

            self.log.warning("Cell with id '%s' exists multiple times!", grade_id)
            cell.metadata = {}
        else:
            self.grade_ids.add(grade_id)

        return cell, resources



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
from http import HTTPStatus

from tornado.web import HTTPError


def parse_ids(*args):
    """
    Transforms loose ids to an id tuple.

    :param args: certain amount of id (int) values
    :return: tuple of ids
    """
    try:
        ids = [int(id) for id in args]
    except ValueError:
        raise HTTPError(HTTPStatus.BAD_REQUEST,
                        reason="All IDs have to be numerical")
    if len(ids) == 1:
        return ids[0]
    return tuple(ids)



"""
Base classes for use by OAuth2 based JupyterHub authenticator classes.

Founded based on work by Kyle Kelley (@rgbkrk)
"""
import base64
import json
import os
import uuid
from urllib.parse import quote, urlencode, urlparse, urlunparse

from .auth import Authenticator
from .crypto import EncryptionUnavailable, InvalidToken, decrypt
from grader_service.handlers.base_handler import BaseHandler
from grader_service.auth.login import LogoutHandler
from grader_service.utils import url_path_join
from tornado import web
from tornado.auth import OAuth2Mixin
from tornado.httpclient import AsyncHTTPClient, HTTPClientError, HTTPRequest
from tornado.httputil import url_concat
from tornado.log import app_log
from traitlets import Any, Bool, Dict, List, Unicode, default


def guess_callback_uri(protocol, host, server_url):
    return f'{protocol}://{host}{url_path_join(server_url, "oauth_callback")}'


STATE_COOKIE_NAME = "grader-oauthenticator-state"


def _serialize_state(state):
    """Serialize OAuth state to a base64 string after passing through JSON"""
    json_state = json.dumps(state)
    return base64.urlsafe_b64encode(json_state.encode("utf8")).decode("ascii")


def _deserialize_state(b64_state):
    """Deserialize OAuth state as serialized in _serialize_state"""
    if isinstance(b64_state, str):
        b64_state = b64_state.encode("ascii")
    try:
        json_state = base64.urlsafe_b64decode(b64_state).decode("utf8")
    except ValueError:
        app_log.error(f"Failed to b64-decode state: {b64_state}")
        return {}
    try:
        return json.loads(json_state)
    except ValueError:
        app_log.error(f"Failed to json-decode state: {json_state}")
        return {}


class OAuthLoginHandler(OAuth2Mixin, BaseHandler):
    """Base class for OAuth login handler

    Typically subclasses will need
    """

    # these URLs are part of the OAuth2Mixin API
    # get them from the Authenticator object
    @property
    def _OAUTH_AUTHORIZE_URL(self):
        return self.authenticator.authorize_url

    @property
    def _OAUTH_ACCESS_TOKEN_URL(self):
        return self.authenticator.token_url

    @property
    def _OAUTH_USERINFO_URL(self):
        return self.authenticator.userdata_url

    def set_state_cookie(self, state):
        self._set_cookie(STATE_COOKIE_NAME, state, expires_days=1.0, httponly=True)

    _state = None

    def get_state(self):
        next_url = original_next_url = self.get_argument("next", None)
        if next_url:
            # avoid browsers treating \ as /
            next_url = next_url.replace("\\", quote("\\"))
            # disallow hostname-having urls,
            # force absolute path redirect
            urlinfo = urlparse(next_url)
            next_url = urlinfo._replace(
                scheme="", netloc="", path="/" + urlinfo.path.lstrip("/")
            ).geturl()
            if next_url != original_next_url:
                self.log.warning(
                    f"Ignoring next_url {original_next_url}, using {next_url}"
                )
        if self._state is None:
            self._state = _serialize_state(
                {"state_id": uuid.uuid4().hex, "next_url": next_url}
            )
        return self._state

    def get(self):
        redirect_uri = self.authenticator.get_callback_url(self)
        token_params = self.authenticator.extra_authorize_params.copy()
        self.log.info(f"OAuth redirect: {redirect_uri}")
        state = self.get_state()
        self.set_state_cookie(state)
        token_params["state"] = state
        self.authorize_redirect(
            redirect_uri=redirect_uri,
            client_id=self.authenticator.client_id,
            scope=self.authenticator.scope,
            extra_params=token_params,
            response_type="code",
        )


class OAuthCallbackHandler(BaseHandler):
    """Basic handler for OAuth callback. Calls authenticator to verify username."""

    _state_cookie = None

    def get_state_cookie(self):
        """Get OAuth state from cookies

        To be compared with the value in redirect URL
        """
        if self._state_cookie is None:
            self._state_cookie = (
                self.get_secure_cookie(STATE_COOKIE_NAME) or b""
            ).decode("utf8", "replace")
            self.clear_cookie(STATE_COOKIE_NAME)
        return self._state_cookie

    def get_state_url(self):
        """Get OAuth state from URL parameters

        to be compared with the value in cookies
        """
        return self.get_argument("state")

    def check_state(self):
        """Verify OAuth state

        compare value in cookie with redirect url param
        """
        cookie_state = self.get_state_cookie()
        url_state = self.get_state_url()
        if not cookie_state:
            raise web.HTTPError(400, reason="OAuth state missing from cookies")
        if not url_state:
            raise web.HTTPError(400, reason="OAuth state missing from URL")
        if cookie_state != url_state:
            self.log.warning(f"OAuth state mismatch: {cookie_state} != {url_state}")
            raise web.HTTPError(400, reason="OAuth state mismatch")

    def check_error(self):
        """Check the OAuth code"""
        error = self.get_argument("error", False)
        if error:
            message = self.get_argument("error_description", error)
            raise web.HTTPError(400, reason=f"OAuth error: {message}")

    def check_code(self):
        """Check the OAuth code"""
        if not self.get_argument("code", False):
            raise web.HTTPError(400, reason="OAuth callback made without a code")

    def check_arguments(self):
        """Validate the arguments of the redirect

        Default:

        - check for oauth-standard error, error_description arguments
        - check that there's a code
        - check that state matches
        """
        self.check_error()
        self.check_code()
        self.check_state()

    def append_query_parameters(self, url, exclude=None):
        """JupyterHub 1.2 appends query parameters by default in get_next_url

        This is not appropriate for oauth callback handlers, where params are oauth state, code, etc.

        Override the method used to append parameters to next_url to not preserve any parameters
        """
        return url

    def get_next_url(self, user=None):
        """Get the redirect target from the state field"""
        state = self.get_state_url()
        if state:
            next_url = _deserialize_state(state).get("next_url")
            if next_url:
                return next_url
        # JupyterHub 0.8 adds default .get_next_url for a fallback
        if hasattr(BaseHandler, "get_next_url"):
            return super().get_next_url(user)
        return self.authenticator.default_next_url

    async def get(self):
        self.check_arguments()
        user = await self.login_user()
        if user is None:
            raise web.HTTPError(403, reason=self.authenticator.custom_403_message)

        self.redirect(self.get_next_url(user))


class OAuthLogoutHandler(LogoutHandler):
    async def handle_logout(self):
        self.clear_cookie(STATE_COOKIE_NAME)

    async def render_logout_page(self):
        if self.authenticator.logout_redirect_url:
            self.redirect(self.authenticator.logout_redirect_url)
            return

        return await super().render_logout_page()


class OAuthenticator(Authenticator):
    """
    Base class for OAuthenticators.

    Subclasses should, in an increasing level of customization:

    - Override the constant `user_auth_state_key`
    - Override various config's default values, such as
      `authorize_url`, `token_url`, `userdata_url`, and `login_service`.
    - Override various methods called by the `authenticate` method, which
      subclasses should not override.
    - Override handler classes such as `login_handler`, `callback_handler`, and
      `logout_handler`.
    """

    login_handler = OAuthLoginHandler
    callback_handler = OAuthCallbackHandler
    logout_handler = OAuthLogoutHandler

    # user_auth_state_key represents the name of the key in the `auth_state`
    # dictionary that user info will be saved
    user_auth_state_key = "oauth_user"

    login_service = Unicode(
        "OAuth 2.0",
        config=True,
        help="""
        Name of the login service or identity provider that this authenticator
        is using to authenticate users.

        This config influences the text on a button shown to unauthenticated
        users before they click it to login, assuming :attr:`auto_login` isn't
        configured True.

        The login button's text will be "Login with <login_service>".
        """,
    )

    allow_all = Bool(
        False,
        config=True,
        help="""
        Allow all authenticated users to login.

        .. versionadded:: 16.0
        """,
    )

    allow_existing_users = Bool(
        False,
        config=True,
        help="""
        Allow existing users to login.

        An existing user is a user in JupyterHub's database of users, and it
        includes all users that has previously logged in.

        .. warning::

           Before enabling this you should review the existing users in the
           JupyterHub admin panel at `/hub/admin`. You may find users existing
           there because they have once been declared in config such as
           `allowed_users` or once been allowed to sign in.

        .. warning::

           When this is enabled and you are to remove access for one or more
           users allowed via other config options, you must make sure that they
           are not part of the database of users still. This can be tricky to do
           if you stop allowing a group of externally managed users for example.

        With this enabled, JupyterHub admin users can visit `/hub/admin` or use
        JupyterHub's REST API to add and remove users as a way to allow them
        access.

        The username for existing users must match the normalized username
        returned by the authenticator. When creating users, only lowercase
        letters should be used unless `MWOAuthenticator` is used.

        .. note::

           Allowing existing users is done by adding existing users on startup
           and newly created users to the `allowed_users` set. Due to that, you
           can't rely on this config to independently allow existing users if
           you for example would reset `allowed_users` after startup.

        .. versionadded:: 16.0

        .. versionchanged:: 16.0

           Before this config was available, the default behavior was to allow
           existing users if `allowed_users` was configured with one or more
           user.
        """,
    )

    authorize_url = Unicode(
        config=True,
        help="""
        The URL to where the user is to be redirected initially based on the
        OAuth2 protocol. The user will be redirected back with an
        `authorization grant code`_ after authenticating successfully with the
        identity provider.

        .. _authorization grant code: https://www.rfc-editor.org/rfc/rfc6749#section-1.3.1

        For more context, see the `Protocol Flow section
        <https://www.rfc-editor.org/rfc/rfc6749#section-1.2>`_ in the OAuth2
        standard document, specifically steps A-B.
        """,
    )

    @default("authorize_url")
    def _authorize_url_default(self):
        return os.environ.get("OAUTH2_AUTHORIZE_URL", "")

    token_url = Unicode(
        config=True,
        help="""
        The URL to where this authenticator makes a request to acquire an
        `access token`_ based on the authorization code received by the user
        returning from the :attr:`authorize_url`.

        .. _access token: https://www.rfc-editor.org/rfc/rfc6749#section-1.4

        For more context, see the `Protocol Flow section
        <https://www.rfc-editor.org/rfc/rfc6749#section-1.2>`_ in the OAuth2
        standard document, specifically steps C-D.
        """,
    )

    @default("token_url")
    def _token_url_default(self):
        return os.environ.get("OAUTH2_TOKEN_URL", "")

    userdata_url = Unicode(
        config=True,
        help="""
        The URL to where this authenticator makes a request to acquire user
        details with an access token received via a request to the
        :attr:`token_url`.

        For more context, see the `Protocol Flow section
        <https://www.rfc-editor.org/rfc/rfc6749#section-1.2>`_ in the OAuth2
        standard document, specifically steps E-F.
        """,
    )

    @default("userdata_url")
    def _userdata_url_default(self):
        return os.environ.get("OAUTH2_USERDATA_URL", "")

    username_claim = Unicode(
        "username",
        config=True,
        help="""
        The key to get the JupyterHub username from in the data response to the
        request made to :attr:`userdata_url`.

        Examples include: email, username, nickname

        What keys are available will depend on the scopes requested and the
        authenticator used.
        """,
    )

    # Enable refresh_pre_spawn by default if self.enable_auth_state
    @default("refresh_pre_spawn")
    def _refresh_pre_spawn_default(self):
        if self.enable_auth_state:
            return True

        return False

    logout_redirect_url = Unicode(
        config=True,
        help="""
        When configured, users are not presented with the JupyterHub logout
        page, but instead redirected to this destination.
        """,
    )

    @default("logout_redirect_url")
    def _logout_redirect_url_default(self):
        return os.getenv("OAUTH_LOGOUT_REDIRECT_URL", "")

    # Originally a GenericOAuthenticator only trait
    userdata_params = Dict(
        config=True,
        help="""
        Userdata params to get user data login information.
        """,
    )

    # Originally a GenericOAuthenticator only trait
    userdata_token_method = Unicode(
        os.environ.get("OAUTH2_USERDATA_REQUEST_TYPE", "header"),
        config=True,
        help="""
        Method for sending access token in userdata request.

        Supported methods: header, url.
        """,
    )

    # Originally a GenericOAuthenticator only trait
    token_params = Dict(
        config=True,
        help="""
        Extra parameters for first POST request exchanging the OAuth code for an
        Access Token
        """,
    )

    custom_403_message = Unicode(
        "Sorry, you are not currently authorized to use this hub. Please contact the hub admin.",
        config=True,
        help="""
        The message to be shown when user was not allowed
        """,
    )

    scope = List(
        Unicode(),
        config=True,
        help="""
        The OAuth scopes to request.

        See the OAuth documentation of your OAuth provider for options.
        """,
    )

    extra_authorize_params = Dict(
        config=True,
        help="""
        Extra GET params to send along with the initial OAuth request to the
        OAuth provider.
        """,
    )

    oauth_callback_url = Unicode(
        os.getenv("OAUTH_CALLBACK_URL", ""),
        config=True,
        help="""
        Callback URL to use.
        
        When registering an OAuth2 application with an identity provider, this
        is typically called the redirect url.

        Should very likely be set to `https://[your-domain]/hub/oauth_callback`.
        """,
    )

    # Originally a GenericOAuthenticator only trait
    basic_auth = Bool(
        os.environ.get("OAUTH2_BASIC_AUTH", "False").lower() in {"true", "1"},
        config=True,
        help="""
        Whether or to use HTTP Basic authentication instead of form based
        authentication in requests to :attr:`token_url`.

        When using HTTP Basic authentication, a HTTP header is set with the
        :attr:`client_id` and :attr:`client_secret` encoded in it.

        When using form based authentication, the `client_id` and
        `client_secret` is put in the HTTP POST request's body.

        .. versionchanged:: 16.0.0

           This configuration now toggles between HTTP Basic authentication and
           form based authentication when working against the `token_url`.

           Previously when this was configured True, both would be used contrary
           to a recommendation in `OAuth 2.0 documentation
           <https://www.rfc-editor.org/rfc/rfc6749#section-2.3.1>`_.

        .. versionchanged:: 16.0.2

           The default value for this configuration for GenericOAuthenticator
           changed from True to False.
        """,
    )

    client_id_env = ""
    client_id = Unicode(
        config=True,
        help="""
        The client id of the OAuth2 application registered with the identity
        provider.
        """,
    )

    def _client_id_default(self):
        if self.client_id_env:
            client_id = os.getenv(self.client_id_env, "")
            if client_id:
                return client_id
        return os.getenv("OAUTH_CLIENT_ID", "")

    client_secret_env = ""
    client_secret = Unicode(
        config=True,
        help="""
        The client secret of the OAuth2 application registered with the identity
        provider.
        """,
    )

    def _client_secret_default(self):
        if self.client_secret_env:
            client_secret = os.getenv(self.client_secret_env, "")
            if client_secret:
                return client_secret
        return os.getenv("OAUTH_CLIENT_SECRET", "")

    validate_server_cert_env = "OAUTH_TLS_VERIFY"
    validate_server_cert = Bool(
        config=True,
        help="""
        Determines if certificates are validated.

        Only set this to False if you feel confident it will not be a security
        concern.
        """,
    )

    def _validate_server_cert_default(self):
        env_value = os.getenv(self.validate_server_cert_env, "")
        if env_value == "0":
            return False
        else:
            return True

    http_request_kwargs = Dict(
        config=True,
        help="""
        Extra default kwargs passed to all HTTPRequests.

        .. code-block:: python

            # Example: send requests through a proxy
            c.OAuthenticator.http_request_kwargs = {
                "proxy_host": "proxy.example.com",
                "proxy_port": 8080,
            }

            # Example: validate against certain root certificates
            c.OAuthenticator.http_request_kwargs = {
                "ca_certs": "/path/to/a.crt",
            }

        See :external:py:class:`tornado.httpclient.HTTPRequest` for all kwargs
        options you can pass. Note that the HTTP client making these requests is
        :external:py:class:`tornado.httpclient.AsyncHTTPClient`.
        """,
    )

    http_client = Any()

    @default("http_client")
    def _default_http_client(self):
        return AsyncHTTPClient()

    default_next_url = Unicode(default_value="http://localhost:8080", config=True,
                               help="""The URL to use when no next URL is specified in the OAuth request.""")

    async def fetch(self, req, label="fetching", parse_json=True, **kwargs):
        """Wrapper for http requests

        logs error responses, parses successful JSON responses

        Args:
            req: tornado HTTPRequest
            label (str): label describing what is happening,
                used in log message when the request fails.
            parse_json (bool): whether to parse the response as JSON
            **kwargs: remaining keyword args
                passed to underlying `client.fetch(req, **kwargs)`
        Returns:
            parsed JSON response if `parse_json=True`, else `tornado.HTTPResponse`
        """
        try:
            resp = await self.http_client.fetch(req, **kwargs)
        except HTTPClientError as e:
            if e.response:
                # Log failed response message for debugging purposes
                message = e.response.body.decode("utf8", "replace")
                try:
                    # guess json, reformat for readability
                    json_message = json.loads(message)
                except ValueError:
                    # not json
                    pass
                else:
                    # reformat json log message for readability
                    message = json.dumps(json_message, sort_keys=True, indent=1)
            else:
                # didn't get a response, e.g. connection error
                message = str(e)

            # log url without query params
            url = urlunparse(urlparse(req.url)._replace(query=""))
            app_log.error(f"Error {label} {e.code} {req.method} {url}: {message}")
            raise e
        else:
            if parse_json:
                if resp.body:
                    return json.loads(resp.body.decode("utf8", "replace"))
                else:
                    # empty body is None
                    return None
            else:
                return resp

    async def httpfetch(
        self, url, label="fetching", parse_json=True, raise_error=True, **kwargs
    ):
        """Wrapper for creating and fetching http requests

        Includes http_request_kwargs in request kwargs
        logs error responses, parses successful JSON responses

        Args:
            url (str): url to fetch
            label (str): label describing what is happening,
                used in log message when the request fails.
            parse_json (bool): whether to parse the response as JSON
            raise_error (bool): whether to raise an exception on HTTP errors
            **kwargs: remaining keyword args
                passed to underlying `tornado.HTTPRequest`, overrides
                `http_request_kwargs`
        Returns:
            parsed JSON response if `parse_json=True`, else `tornado.HTTPResponse`
        """
        request_kwargs = self.http_request_kwargs.copy()
        request_kwargs.update(kwargs)
        req = HTTPRequest(url, **request_kwargs)
        return await self.fetch(
            req, label=label, parse_json=parse_json, raise_error=raise_error
        )

    def add_user(self, user):
        """
        Overrides `Authenticator.add_user`, a hook called for all users in the
        database on startup and for each user being created.

        The purpose of the override is to implement the `allow_existing_users`
        config by adding users to the `allowed_users` set only if
        `allow_existing_users` is truthy. The overridden behavior is to do it if
        `allowed_users` is truthy.

        The implementation is adjusted from JupyterHub 4.0.1:
        https://github.com/jupyterhub/jupyterhub/blob/4.0.1/jupyterhub/auth.py#L625-L648
        """
        if not self.validate_username(user.name):
            raise ValueError("Invalid username: %s" % user.name)
        if not self.allow_all and self.allow_existing_users:
            self.allowed_users.add(user.name)

    def login_url(self, base_url):
        return url_path_join(base_url, "oauth_login")

    def logout_url(self, base_url):
        return url_path_join(base_url, "logout")

    def get_callback_url(self, handler=None):
        """Get my OAuth redirect URL

        Either from config or guess based on the current request.
        """
        if self.oauth_callback_url:
            return self.oauth_callback_url
        elif handler:
            return guess_callback_uri(
                handler.request.protocol,
                handler.request.host,
                handler.application.base_url,
            )
        else:
            raise ValueError(
                "Specify callback oauth_callback_url or give me a handler to guess with"
            )

    def get_handlers(self, base_url_path):
        self.log.info(base_url_path)
        return [
            (f"{base_url_path}/oauth_login", self.login_handler),
            (f"{base_url_path}/oauth_callback", self.callback_handler),
            (f"{base_url_path}/logout", self.logout_handler),
        ]

    def build_userdata_request_headers(self, access_token, token_type):
        """
        Builds and returns the headers to be used in the userdata request.
        Called by the :meth:`oauthenticator.OAuthenticator.token_to_user`
        """
        return {
            "Accept": "application/json",
            "User-Agent": "JupyterHub",
            "Authorization": f"{token_type} {access_token}",
        }

    def build_token_info_request_headers(self):
        """
        Builds and returns the headers to be used in the access token request.
        Called by the :meth:`oauthenticator.OAuthenticator.get_token_info`.

        The Content-Type header is specified by the OAuth 2.0 RFC in
        https://www.rfc-editor.org/rfc/rfc6749#section-4.1.3. utf-8 is also
        required according to https://www.rfc-editor.org/rfc/rfc6749#appendix-B,
        and that can be specified with a Content-Type directive according to
        https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Type#directives.
        """
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/x-www-form-urlencoded; charset=utf-8",
            "User-Agent": "JupyterHub",
        }

        if self.basic_auth:
            b64key = base64.b64encode(
                bytes("{self.client_id}:{self.client_secret}", "utf8")
            )
            headers.update({"Authorization": f'Basic {b64key.decode("utf8")}'})
        return headers

    def user_info_to_username(self, user_info):
        """
        Gets the self.username_claim key's value from the user_info dictionary.

        Should be overridden by the authenticators for which the hub username cannot
        be extracted this way and needs extra processing.

        Args:
            user_info: the dictionary returned by the userdata request

        Returns:
            user_info["self.username_claim"] or raises an error if such value isn't found.

        Called by the :meth:`oauthenticator.OAuthenticator.authenticate`
        """
        username = user_info.get(self.username_claim, None)
        if not username:
            message = (f"No {self.username_claim} found in {user_info}",)
            self.log.error(message)
            raise ValueError(message)

        return username

    # Originally a GoogleOAuthenticator only feature
    async def get_prev_refresh_token(self, handler, username):
        """
        Retrieves the `refresh_token` from previous encrypted auth state.
        Called by the :meth:`oauthenticator.OAuthenticator.authenticate`
        """
        user = handler.find_user(username)
        if not user or not user.encrypted_auth_state:
            return

        self.log.debug(
            "Encrypted_auth_state was found, will try to decrypt and pull refresh_token from it..."
        )

        try:
            encrypted = user.encrypted_auth_state
            auth_state = await decrypt(encrypted)

            return auth_state.get("refresh_token")
        except (ValueError, InvalidToken, EncryptionUnavailable) as e:
            self.log.warning(
                f"Failed to retrieve encrypted auth_state for {username}. Error was {e}.",
            )
            return

    def build_access_tokens_request_params(self, handler, data=None):
        """
        Builds the parameters that should be passed to the URL request
        that exchanges the OAuth code for the Access Token.
        Called by the :meth:`oauthenticator.OAuthenticator.authenticate`.
        """
        code = handler.get_argument("code")
        if not code:
            raise web.HTTPError(400, reason="Authentication Cancelled.")

        params = {
            "code": code,
            "grant_type": "authorization_code",
            "redirect_uri": self.get_callback_url(handler),
            "data": data,
        }

        # the client_id and client_secret should not be included in the access token request params
        # when basic authentication is used
        # ref: https://www.rfc-editor.org/rfc/rfc6749#section-2.3.1
        if not self.basic_auth:
            params.update(
                [("client_id", self.client_id), ("client_secret", self.client_secret)]
            )

        params.update(self.token_params)

        return params

    async def get_token_info(self, handler, params):
        """
        Makes a "POST" request to `self.token_url`, with the parameters received as argument.

        Returns:
            the JSON response to the `token_url` the request.

        Called by the :meth:`oauthenticator.OAuthenticator.authenticate`
        """

        token_info = await self.httpfetch(
            self.token_url,
            method="POST",
            headers=self.build_token_info_request_headers(),
            body=urlencode(params).encode("utf-8"),
            validate_cert=self.validate_server_cert,
        )

        if "error_description" in token_info:
            raise web.HTTPError(
                403,
                f'An access token was not returned: {token_info["error_description"]}',
            )
        elif "access_token" not in token_info:
            raise web.HTTPError(500, reason=f"Bad response: {token_info}")

        return token_info

    async def token_to_user(self, token_info):
        """
        Determines who the logged-in user by sending a "GET" request to
        :data:`oauthenticator.OAuthenticator.userdata_url` using the `access_token`.

        Args:
            token_info: the dictionary returned by the token request (exchanging the OAuth code for an Access Token)

        Returns:
            the JSON response to the `userdata_url` request.

        Called by the :meth:`oauthenticator.OAuthenticator.authenticate`
        """
        access_token = token_info["access_token"]
        token_type = token_info["token_type"]

        if not self.userdata_url:
            raise ValueError(
                "authenticator.userdata_url is missing. Please configure it."
            )

        url = url_concat(self.userdata_url, self.userdata_params)
        if self.userdata_token_method == "url":
            url = url_concat(url, dict(access_token=access_token))

        return await self.httpfetch(
            url,
            "Fetching user info...",
            method="GET",
            headers=self.build_userdata_request_headers(access_token, token_type),
            validate_cert=self.validate_server_cert,
        )

    def build_auth_state_dict(self, token_info, user_info):
        """
        Builds the `auth_state` dict that will be returned by a succesfull `authenticate` method call.

        Args:
            token_info: the dictionary returned by the token request (exchanging the OAuth code for an Access Token)
            user_info: the dictionary returned by the userdata request

        Returns:
            auth_state: a dictionary of auth state that should be persisted with the following keys:
                - "access_token": the access_token
                - "refresh_token": the refresh_token, if available
                - "id_token": the id_token, if available
                - "scope": the scopes, if available
                - "token_response": the full token_info response
                - self.user_auth_state_key: the full user_info response

        Called by the :meth:`oauthenticator.OAuthenticator.authenticate`
        """

        # We know for sure the `access_token` key exists, oterwise we would have errored out already
        access_token = token_info["access_token"]

        refresh_token = token_info.get("refresh_token", None)
        id_token = token_info.get("id_token", None)
        scope = token_info.get("scope", "")

        if isinstance(scope, str):
            scope = scope.split(" ")

        return {
            "access_token": access_token,
            "refresh_token": refresh_token,
            "id_token": id_token,
            "scope": scope,
            # Save the full token response
            # These can be used for user provisioning in the Lab/Notebook environment.
            "token_response": token_info,
            # store the whole user model in auth_state too
            self.user_auth_state_key: user_info,
        }

    async def update_auth_model(self, auth_model):
        """
        Updates and returns the `auth_model` dict.

        Should be overridden to collect information required for check_allowed.

        Args: auth_model - the auth model dictionary, containing:
            - `name`: the normalized username
            - `admin`: the admin status (True/False/None), where None means it
                should be unchanged.
            - `auth_state`: the dictionary of of auth state
                returned by :meth:`oauthenticator.OAuthenticator.build_auth_state_dict`

        Called by the :meth:`oauthenticator.OAuthenticator.authenticate`
        """
        return auth_model

    async def authenticate(self, handler, data=None, **kwargs):
        """
        A JupyterHub Authenticator's authenticate method's job is:

        - return None if the user isn't successfully authenticated
        - return a dictionary if authentication is successful with name, admin
          (optional), and auth_state (optional)

        Subclasses should not override this method.

        ref: https://jupyterhub.readthedocs.io/en/stable/reference/authenticators.html#authenticator-authenticate-method
        ref: https://github.com/jupyterhub/jupyterhub/blob/4.0.0/jupyterhub/auth.py#L581-L611
        """
        # build the parameters to be used in the request exchanging the oauth code for the access token
        access_token_params = self.build_access_tokens_request_params(handler, data)
        # exchange the oauth code for an access token and get the JSON with info about it
        token_info = await self.get_token_info(handler, access_token_params)
        # use the access_token to get userdata info
        user_info = await self.token_to_user(token_info)
        # extract the username out of the user_info dict and normalize it
        username = self.user_info_to_username(user_info)
        username = self.normalize_username(username)

        # check if there any refresh_token in the token_info dict
        refresh_token = token_info.get("refresh_token", None)
        if self.enable_auth_state and not refresh_token:
            self.log.debug(
                "Refresh token was empty, will try to pull refresh_token from previous auth_state"
            )
            refresh_token = await self.get_prev_refresh_token(handler, username)
            if refresh_token:
                token_info["refresh_token"] = refresh_token

        # build the auth model to be read if authentication goes right
        auth_model = {
            "name": username,
            "admin": True if username in self.admin_users else None,
            "auth_state": self.build_auth_state_dict(token_info, user_info),
        }

        # update the auth_model with info to later authorize the user in
        # check_allowed, such as admin status and group memberships
        return await self.update_auth_model(auth_model)

    async def check_allowed(self, username, auth_model):
        """
        Returns True for users allowed to be authorized

        Overrides Authenticator.check_allowed that is called from
        `Authenticator.get_authenticated_user` after
        `OAuthenticator.authenticate` has been called, and therefore also after
        `update_auth_model` has been called.

        Subclasses with additional config to allow a user should override this
        method and return True when this method returns True or if a user is
        allowed via the additional config.
        """
        # A workaround for JupyterHub<=4.0.1, described in
        # https://github.com/jupyterhub/oauthenticator/issues/621
        if auth_model is None:
            return True

        if self.allow_all:
            return True

        # allow users with admin status set to True via admin_users config or
        # update_auth_model override
        if auth_model["admin"]:
            return True

        # allow users in allowed_users, note that allowed_users is appended
        # automatically with existing users if it was configured truthy
        if username in self.allowed_users:
            return True

        # users should be explicitly allowed via config, otherwise they aren't
        return False

    # _deprecated_oauth_aliases should be a dictionary with a format as:
    #
    # {
    #     "old_config": (
    #         "new_config",
    #         "16.0.0",      # version when it became deprecated
    #         False,         # if new config can be updated with a warning
    #     ),
    # }
    #
    _deprecated_oauth_aliases = {}

    def _deprecated_oauth_trait(self, change):
        """observer for deprecated traits"""
        old_attr = change.name
        try:
            new_attr, version, same = self._deprecated_oauth_aliases.get(old_attr)
        except ValueError:
            # if `same` flag wasn't passed, we assume the new and old trait have the same type
            new_attr, version = self._deprecated_oauth_aliases.get(old_attr)
            same = True

        new_value = getattr(self, new_attr)
        if new_value != change.new:
            # only warn if different
            # protects backward-compatible config from warnings
            # if they set the same value under both names
            message = "{cls}.{old} is deprecated in {cls} {version}, use {cls}.{new} instead".format(
                cls=self.__class__.__name__,
                old=old_attr,
                new=new_attr,
                version=version,
            )

            # set the value for the new attr only if they are the same type
            # otherwise raise an error because unexpected things can happen
            if same:
                self.log.warning(message)
                setattr(self, new_attr, change.new)
            else:
                self.log.error(message)
                raise ValueError(message)

    def __init__(self, **kwargs):
        # observe deprecated config names in oauthenticator
        if self._deprecated_oauth_aliases:
            self.observe(
                self._deprecated_oauth_trait, names=list(self._deprecated_oauth_aliases)
            )
        super().__init__(**kwargs)



"""Add feedback status

Revision ID: 9cfeb0faa0c0
Revises: 4620c13728e7
Create Date: 2024-01-10 15:16:36.273796

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '9cfeb0faa0c0'
down_revision = '4620c13728e7'
branch_labels = None
depends_on = None


def upgrade():
    sa_enum = sa.Enum('not_generated', 'generated', 'generating', 'generation_failed', 'feedback_outdated', name='feedback_status')
    # This will create the enum if it's not already present
    sa_enum.create(op.get_bind(), checkfirst=True)
    op.add_column('submission', sa.Column('feedback_status', sa_enum, server_default='not_generated', nullable=False))

    op.execute("UPDATE submission SET feedback_status='not_generated' WHERE feedback_available=false")
    op.execute("UPDATE submission SET feedback_status='generated' WHERE feedback_available=true")
    op.drop_column('submission', 'feedback_available')

def downgrade():
    op.add_column('submission', sa.Column('feedback_available', sa.Boolean, server_default='false', nullable=False))
    op.execute("UPDATE submission SET feedback_available=false WHERE feedback_status='not_generated' OR feedback_status='generation_failed' OR feedback_status='generating'")
    op.execute("UPDATE submission SET feedback_available=true WHERE feedback_status='generated' OR feedback_status='feedback_outdated'")
    op.drop_column('submission', 'feedback_status')

    sa_enum = sa.Enum(name='feedback_status')
    sa_enum.drop(op.get_bind(), checkfirst=True)




# Default Grader Convert Configuration File

# -------------------------------------------------
# ClearSolutions Configuration
# -------------------------------------------------

# `code_stub` replaces the content of solution cells with a language-specific code snippet.
# Instructors can override these placeholders with the code snippets of their choice.
# c.ClearSolutions.code_stub = {
    # 'python': "# YOUR CODE HERE\nraise NotImplementedError()",  # Placeholder for Python solution cells
    # 'matlab': "% YOUR CODE HERE\nerror('No Answer Given!')",  # Placeholder for MATLAB solution cells
    # 'octave': "% YOUR CODE HERE\nerror('No Answer Given!')",  # Placeholder for Octave solution cells
    # ... (More languages)
# }

# -------------------------------------------------
# ClearAlwaysHiddenTests Configuration
# -------------------------------------------------

# Delimiters for always hidden utilities.
# Blocks of code between these delimiters will always be hidden in the notebook.
# c.ClearAlwaysHiddenTests.begin_util_delimeter = "BEGIN ALWAYS HIDDEN TESTS"  # Start delimiter
# c.ClearAlwaysHiddenTests.end_util_delimeter = "END ALWAYS HIDDEN TESTS"  # End delimiter

# -------------------------------------------------
# ClearHiddenTests Configuration
# -------------------------------------------------

# Delimiters for hidden tests.
# Blocks of code between these delimiters will be hidden in the notebook.
# c.ClearHiddenTests.begin_test_delimeter = "BEGIN HIDDEN TESTS"  # Start delimiter
# c.ClearHiddenTests.end_test_delimeter = "END HIDDEN TESTS"  # End delimiter

# -------------------------------------------------
# ClearMarkScheme Configuration
# -------------------------------------------------

# Delimiters for the mark scheme.
# Blocks of text between these delimiters describe the marking scheme.
# c.ClearMarkScheme.begin_mark_scheme_delimeter = "BEGIN MARK SCHEME"  # Start delimiter
# c.ClearMarkScheme.end_mark_scheme_delimeter = "END MARK SCHEME"  # End delimiter

# -------------------------------------------------
# IncludeHeaderFooter Configuration
# -------------------------------------------------

# Header and Footer files to be included at the top and bottom of each notebook.
# c.IncludeHeaderFooter.header = "../../source/<assignment_id>/header.ipynb"  # Header notebook file
# c.IncludeHeaderFooter.footer = "../../source/<assignment_id>/footer.ipynb"  # Footer notebook file

# -------------------------------------------------
# LimitOutput Configuration
# -------------------------------------------------

# Limit the number of lines and traceback lines in the output cells.
# c.LimitOutput.max_lines = 1000  # Max number of lines in output
# c.LimitOutput.max_traceback = 100  # Max number of traceback lines

# -------------------------------------------------
# LockCells Configuration
# -------------------------------------------------

# Options for locking cells in the notebook to prevent editing.
# c.LockCells.lock_solution_cells = True  # Lock solution cells
# c.LockCells.lock_grade_cells = True  # Lock grade cells
# c.LockCells.lock_readonly_cells = True  # Lock readonly cells
# c.LockCells.lock_all_cells = False  # Lock all cells in the notebook (overrides above settings)

# -------------------------------------------------
# END OF CONFIGURATION FILE
# -------------------------------------------------



from grader_service.handlers.base_handler import (GraderBaseHandler,
                                                  RequestHandlerConfig,
                                                  authorize)
from grader_service.orm.takepart import Scope
from grader_service.registry import register_handler, VersionSpecifier


@register_handler(
    path=r"\/api\/config\/?",
    version_specifier=VersionSpecifier.ALL,
)
class ConfigHandler(GraderBaseHandler):
    """
    Handler class for requests to /config
    """

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self):
        """
        Gathers useful config for the grader labextension and returns it.
        :return: config in dict
        """
        self.write({})



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# import handlers
from grader_service.handlers import (assignment, grading, lectures,
                                     submissions, git,
                                     permission, health, config, base_handler)

__all__ = ["assignment", "grading", "lectures", "submissions",
           "git", "permission", "health", "config", "base_handler"]



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import io
import logging
import os
import shutil
from subprocess import CalledProcessError

from traitlets import Unicode

from grader_service.autograding.local_grader import (LocalAutogradeExecutor,
                                                     rm_error)
from grader_service.orm.assignment import Assignment
from grader_service.orm.group import Group
from grader_service.orm.lecture import Lecture
from grader_service.orm.submission import Submission
from grader_service.convert.converters.generate_feedback import GenerateFeedback


class GenerateFeedbackExecutor(LocalAutogradeExecutor):
    def __init__(self, grader_service_dir: str,
                 submission: Submission, **kwargs):
        super().__init__(grader_service_dir, submission, **kwargs)

    @property
    def input_path(self):
        return os.path.join(self.grader_service_dir, self.relative_input_path,
                            f"feedback_{self.submission.id}")

    @property
    def output_path(self):
        return os.path.join(self.grader_service_dir, self.relative_output_path,
                            f"feedback_{self.submission.id}")

    def _pull_submission(self):
        if not os.path.exists(self.input_path):
            os.mkdir(self.input_path)

        assignment: Assignment = self.submission.assignment
        lecture: Lecture = assignment.lecture

        if assignment.type == "user":
            repo_name = self.submission.username
        else:
            group = self.session.query(Group).get(
                (self.submission.username, lecture.id)
            )
            if group is None:
                raise ValueError()
            repo_name = group.name

        git_repo_path = os.path.join(
            self.grader_service_dir,
            "git",
            lecture.code,
            str(assignment.id),
            "autograde",
            assignment.type,
            repo_name,
        )

        if os.path.exists(self.input_path):
            shutil.rmtree(self.input_path, onerror=rm_error)
        os.mkdir(self.input_path)

        self.log.info(f"Pulling repo {git_repo_path} into input directory")

        command = f"{self.git_executable} init"
        self.log.info(f"Running {command}")
        try:
            self._run_subprocess(command, self.input_path)
        except CalledProcessError:
            pass

        command = f'{self.git_executable} pull "{git_repo_path}" ' \
                  f'submission_{self.submission.commit_hash}'
        self.log.info(f"Running {command}")
        try:
            self._run_subprocess(command, self.input_path)
        except CalledProcessError:
            pass
        self.log.info("Successfully cloned repo")

    def _run(self):
        if os.path.exists(self.output_path):
            shutil.rmtree(self.output_path, onerror=rm_error)

        os.makedirs(self.output_path, exist_ok=True)
        self._write_gradebook(self.submission.properties.properties)

        autograder = GenerateFeedback(self.input_path, self.output_path,
                                      "*.ipynb", copy_files=False)
        autograder.force = True

        log_stream = io.StringIO()
        log_handler = logging.StreamHandler(log_stream)
        autograder.log.addHandler(log_handler)

        autograder.start()

        self.grading_logs = log_stream.getvalue()
        autograder.log.removeHandler(log_handler)

    def _push_results(self):
        os.unlink(os.path.join(self.output_path, "gradebook.json"))

        assignment: Assignment = self.submission.assignment
        lecture: Lecture = assignment.lecture

        if assignment.type == "user":
            repo_name = self.submission.username
        else:
            group = self.session.query(Group).get(
                (self.submission.username, lecture.id)
            )
            if group is None:
                raise ValueError()
            repo_name = group.name

        git_repo_path = os.path.join(
            self.grader_service_dir,
            "git",
            lecture.code,
            str(assignment.id),
            "feedback",
            assignment.type,
            repo_name,
        )

        if not os.path.exists(git_repo_path):
            os.makedirs(git_repo_path, exist_ok=True)
            try:
                self._run_subprocess(
                    f'git init --bare "{git_repo_path}"', self.output_path
                )
            except CalledProcessError:
                raise

        command = f"{self.git_executable} init"
        self.log.info(f"Running {command} at {self.output_path}")
        try:
            self._run_subprocess(command, self.output_path)
        except CalledProcessError:
            pass

        self.log.info(f"Creating new branch "
                      f"feedback_{self.submission.commit_hash}")
        command = (
            f"{self.git_executable} switch -c "
            f"feedback_{self.submission.commit_hash}"
        )
        try:
            self._run_subprocess(command, self.output_path)
        except CalledProcessError:
            pass
        self.log.info(f"Now at branch "
                      f"feedback_{self.submission.commit_hash}")

        self.log.info(f"Commiting all files in {self.output_path}")
        self._run_subprocess(
            f"{self.git_executable} add -A", self.output_path
        )
        self._run_subprocess(
            f'{self.git_executable} commit -m "{self.submission.commit_hash}"',
            self.output_path,
        )
        self.log.info(
            f"Pushing to {git_repo_path} at branch "
            f"feedback_{self.submission.commit_hash}"
        )
        command = f'{self.git_executable} push -uf "{git_repo_path}" ' \
                  f'feedback_{self.submission.commit_hash}'
        self._run_subprocess(command, self.output_path)
        self.log.info("Pushing complete")

    def _set_properties(self):
        # No need to set properties
        pass

    def _set_db_state(self, success=True):
        """"
        Sets the submission feedback status based on the success of the generation.
        :param success: Whether feedback generation was succesfull or not.
        :return: None
        """
        if success:
            self.submission.feedback_status = "generated"
        else:
            self.submission.feedback_status = "generation_failed"
        self.session.commit()


class GenerateFeedbackProcessExecutor(GenerateFeedbackExecutor):
    convert_executable = Unicode("grader-convert",
                                 allow_none=False).tag(config=True)

    def _run(self):
        if os.path.exists(self.output_path):
            shutil.rmtree(self.output_path, onerror=rm_error)

        os.mkdir(self.output_path)
        self._write_gradebook(self.submission.properties)

        command = f'{self.convert_executable} generate_feedback -i ' \
                  f'"{self.input_path}" -o "{self.output_path}" -p "*.ipynb"'
        self.log.info(f"Running {command}")
        process = self._run_subprocess(command, None)
        self.grading_logs = process.stderr.read().decode("utf-8")
        self.log.info(self.grading_logs)
        if process.returncode == 0:
            self.log.info("Process has successfully completed execution!")
        else:
            raise RuntimeError("Process has failed execution!")



"""Settings column

Revision ID: 4620c13728e7
Revises: a1791b1371ed
Create Date: 2023-09-11 15:46:57.162396

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '4620c13728e7'
down_revision = 'a1791b1371ed'
branch_labels = None
depends_on = None


def upgrade():
    # use settings to save arbitrary json
    op.add_column('assignment', sa.Column('settings', sa.Text, server_default='{}', nullable=False))

    # columns for original points (grading_points) and the late submission scaling (point_scaling)
    op.add_column('submission', sa.Column('grading_score', sa.DECIMAL(10, 3), nullable=True))
    op.add_column('submission', sa.Column('score_scaling', sa.DECIMAL(10, 3), server_default="1.0", nullable=False))


def downgrade():
    op.drop_column('assignment', 'settings')
    op.drop_column('submission', 'grading_score')
    op.drop_column('submission', 'score_scaling')



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# Python script that will apply the migrations up to head
import argparse
import logging
import alembic.config
import os
import re
from urllib.parse import urlparse, urlunparse

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Path to the directory containing this script
script_dir = os.path.dirname(os.path.abspath(__file__))

# Alembic command arguments
alembic_args = [
    '-c', os.path.join(script_dir, 'alembic.ini'),
    'upgrade', 'head'
]

def redact_db_url(url):
    """Redact sensitive parts of the database URL for logging."""
    parsed_url = urlparse(url)
    netloc = parsed_url.netloc

    if '@' in netloc:
        userinfo, hostinfo = netloc.split('@', 1)
        username = re.sub(r':.*$', '', userinfo)  # Extract username
        netloc = f"{username}:REDACTED@{hostinfo}"
    
    redacted_url = urlunparse((
        parsed_url.scheme,
        netloc,
        parsed_url.path,
        parsed_url.params,
        parsed_url.query,
        parsed_url.fragment
    ))
    return redacted_url

def get_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description="Apply Alembic migrations up to head.")
    parser.add_argument("--config", "-f", default=None, help="Path to custom config file")
    return parser.parse_args()

def load_custom_config(config_path):
    """Load and apply configuration from a custom file."""
    from grader_service.main import GraderService

    try:
        service = GraderService.instance()
        service.load_config_file(config_path)
        service.set_config()
        db_url = service.db_url
        logger.info(f'Database URL: {redact_db_url(db_url)}')
        
        if db_url:
            os.environ["GRADER_DB_URL"] = db_url
        else:
            logger.info('No db_url configuration found, using default db_url')
    except Exception as e:
        logger.error(f"Failed to load custom config: {e}")
        raise

def main():
    args = get_args()
    if args.config:
        load_custom_config(args.config)
    
    try:
        alembic.config.main(argv=alembic_args)
    except SystemExit as e:
        logger.error(f"Migration failed: {e}")
        raise

if __name__ == '__main__':
    main()




from textwrap import dedent
from typing import Tuple

from nbconvert.exporters.exporter import ResourcesDict
from nbformat.notebooknode import NotebookNode
from traitlets import Bool, Unicode

from grader_service.convert import utils
from grader_service.convert.preprocessors.base import NbGraderPreprocessor


class ClearHiddenTests(NbGraderPreprocessor):

    begin_test_delimeter = Unicode(
        "BEGIN HIDDEN TESTS",
        help="The delimiter marking the beginning of hidden tests cases",
    ).tag(config=True)

    end_test_delimeter = Unicode(
        "END HIDDEN TESTS", help="The delimiter marking the end of hidden tests cases"
    ).tag(config=True)

    enforce_metadata = Bool(
        True,
        help=dedent(
            """
            Whether or not to complain if cells containing hidden test regions
            are not marked as grade cells. WARNING: this will potentially cause
            things to break if you are using the full nbgrader pipeline. ONLY
            disable this option if you are only ever planning to use nbgrader
            assign.
            """
        ),
    ).tag(config=True)

    def _remove_hidden_test_region(self, cell: NotebookNode) -> bool:
        """Find a region in the cell that is delimeted by
        `self.begin_test_delimeter` and `self.end_test_delimeter` (e.g.  ###
        BEGIN HIDDEN TESTS and ### END HIDDEN TESTS). Remove that region
        depending the cell type.

        This modifies the cell in place, and then returns True if a
        hidden test region was removed, and False otherwise.
        """
        # pull out the cell input/source
        lines = cell.source.split("\n")

        new_lines = []
        in_test = False
        removed_test = False

        for line in lines:
            # begin the test area
            if self.begin_test_delimeter in line:

                # check to make sure this isn't a nested BEGIN HIDDEN TESTS
                # region
                if in_test:
                    raise RuntimeError(
                        "Encountered nested begin hidden tests statements"
                    )
                in_test = True
                removed_test = True

            # end the solution area
            elif self.end_test_delimeter in line:
                in_test = False

            # add lines as long as it's not in the hidden tests region
            elif not in_test:
                new_lines.append(line)

        # we finished going through all the lines, but didn't find a
        # matching END HIDDEN TESTS statment
        if in_test:
            raise RuntimeError("No end hidden tests statement found")

        # replace the cell source
        cell.source = "\n".join(new_lines)

        return removed_test

    def preprocess(
        self, nb: NotebookNode, resources: ResourcesDict
    ) -> Tuple[NotebookNode, ResourcesDict]:
        nb, resources = super(ClearHiddenTests, self).preprocess(nb, resources)
        if "celltoolbar" in nb.metadata:
            del nb.metadata["celltoolbar"]
        return nb, resources

    def preprocess_cell(
        self, cell: NotebookNode, resources: ResourcesDict, cell_index: int
    ) -> Tuple[NotebookNode, ResourcesDict]:
        # remove hidden test regions
        removed_test = self._remove_hidden_test_region(cell)

        # determine whether the cell is a grade cell
        is_grade = utils.is_grade(cell)

        # check that it is marked as a grade cell if we remove a test
        # region -- if it's not, then this is a problem, because the cell needs
        # to be given an id
        if not is_grade and removed_test:
            if self.enforce_metadata:
                raise RuntimeError(
                    "Hidden test region detected in a non-grade cell; "
                    "please make sure all solution regions are within "
                    "'Autograder tests' cells."
                )

        return cell, resources





# coding: utf-8

import glob
import traceback

from traitlets import default

from grader_service.convert.nbgraderformat import SchemaTooNewError, SchemaTooOldError
from grader_service.convert.validator import Validator
from grader_service.convert.converters.baseapp import ConverterApp

aliases = {}
flags = {
    "invert": (
        {"Validator": {"invert": True}},
        "Complain when cells pass, rather than vice versa.",
    )
}


class ValidateApp(ConverterApp):

    name = u"validate"
    description = u"Validate a notebook by running it"

    aliases = aliases
    flags = flags

    examples = """
        You can run `grader-convert validate` on just a single file, e.g.:
            grader-convert validate "Problem 1.ipynb"

        Or, you can run it on multiple files using shell globs:
            grader-convert validate "Problem Set 1/*.ipynb"

        If you want to test instead that none of the tests pass (rather than that
        all of the tests pass, which is the default), you can use --invert:
            grader-convert validate --invert "Problem 1.ipynb"
        """

    @default("classes")
    def _classes_default(self):
        classes = super(ValidateApp, self)._classes_default()
        classes.append(Validator)
        return classes

    def start(self):
        if not len(self.extra_args):
            self.fail("Must provide path to notebook:\nnbgrader validate NOTEBOOK")
        else:
            notebook_filenames = []
            for x in self.extra_args:
                notebook_filenames.extend(glob.glob(x))

        validator = Validator(parent=self)
        for filename in notebook_filenames:
            try:
                validator.validate_and_print(filename)

            except SchemaTooOldError:
                self.log.error(traceback.format_exc())
                self.fail(
                    (
                        "The notebook '{}' uses an old version "
                        "of the nbgrader metadata format. Please **back up this "
                        "notebook** and then update the metadata using:\n\nnbgrader update {}\n"
                    ).format(filename, filename)
                )

            except SchemaTooNewError:
                self.log.error(traceback.format_exc())
                self.fail(
                    (
                        "The notebook '{}' uses a newer version "
                        "of the nbgrader metadata format. Please update your version of "
                        "nbgrader to the latest version to be able to use this notebook."
                    ).format(filename)
                )

            except Exception:
                self.log.error(traceback.format_exc())
                self.fail(
                    "nbgrader encountered a fatal error while trying to validate '{}'".format(
                        filename
                    )
                )
